from __future__ import annotations
from typing import TYPE_CHECKING
import logging

import claripy

from cle.backends.elf.compilation_unit import CompilationUnit
from cle.backends.elf.variable import Variable
from cle.backends.elf.elf import ELF

from .plugin import KnowledgeBasePlugin

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase

l = logging.getLogger(name=__name__)


class DebugVariableContainer:
    """
    Variable tree for variables with same name to lock up which variable is visible at a given program counter address.
    """

    def __init__(self):
        """
        It is recommended to use DebugVariableManager.add_variable() instead
        """
        self.less_visible_vars = []

    def _insertvar(self, var: DebugVariable):
        for i, v in enumerate(self.less_visible_vars):
            if var.test_unsupported_overlap(v):
                if var.cle_variable.declaration_only:
                    # ignore var
                    return
                if v.cle_variable.declaration_only:
                    # ignore v
                    self.less_visible_vars[i] = var
                    var.less_visible_vars = v.less_visible_vars
                    return
                l.warning(
                    'Unsupported variable with overlapping scopes. Have "%s" with %d-%d and ignore %d-%d.',
                    v.cle_variable.name,
                    v.low_pc,
                    v.high_pc,
                    var.low_pc,
                    var.high_pc,
                )
                return
            if var.contains(v):
                self.less_visible_vars[i] = var
                var.less_visible_vars.append(v)
                return
            if v.contains(var):
                v._insertvar(var)
                return
        self.less_visible_vars.append(var)

    def __setitem__(self, index, value):
        assert isinstance(index, slice) and isinstance(value, Variable)
        low_pc = index.start
        high_pc = index.stop
        dvar = DebugVariable(low_pc, high_pc, value)
        return self._insertvar(dvar)

    def from_pc(self, pc) -> Variable:
        """
        Returns the visible variable (if any) for a given pc address.
        """
        for var in self.less_visible_vars:
            if claripy.is_true(var.low_pc <= pc) and claripy.is_true(pc < var.high_pc):
                return var.from_pc(pc)
        return None

    def __getitem__(self, index):
        return self.from_pc(index)


class DebugVariable(DebugVariableContainer):
    """
    :ivar low_pc:           Start of the visibility scope of the variable as program counter address (rebased)
    :ivar high_pc:          End of the visibility scope of the variable as program counter address (rebased)
    :ivar cle_variable:     Original variable from cle
    """

    def __init__(self, low_pc: int, high_pc: int, cle_variable: Variable):
        """
        It is recommended to use DebugVariableManager.add_variable() instead
        """
        super().__init__()
        self.low_pc = low_pc
        self.high_pc = high_pc
        self.cle_variable = cle_variable

    # overwrites the method of DebugVariableContainer
    def from_pc(self, pc) -> Variable:
        if claripy.is_true(pc < self.low_pc) or claripy.is_true(self.high_pc < pc):
            # not within range
            return None
        for var in self.less_visible_vars:
            if claripy.is_true(var.low_pc <= pc) and claripy.is_true(pc < var.high_pc):
                return var.from_pc(pc)
        return self.cle_variable

    def contains(self, dvar: DebugVariable) -> bool:
        return self.low_pc <= dvar.low_pc and dvar.high_pc <= self.high_pc

    def test_unsupported_overlap(self, dvar: DebugVariable) -> bool:
        """
        Test for an unsupported overlapping

        :param dvar:    Second DebugVariable to compare with
        :return:        True if there is an unsupported overlapping
        """
        l1 = self.low_pc
        l2 = dvar.low_pc
        h1 = self.high_pc
        h2 = dvar.high_pc
        if l1 == l2 and h1 == h2:
            return True
        if l2 < l1 < h2 < h1:
            return True
        return l1 < l2 < h1 < h2


class DebugVariableManager(KnowledgeBasePlugin):
    """
    Structure to manage and access variables with different visibility scopes.
    """

    def __init__(self, kb: KnowledgeBase):
        super().__init__(kb=kb)
        self._dvar_containers: dict[str, DebugVariableContainer] = {}

    def from_name_and_pc(self, var_name: str, pc_addr: int) -> Variable:
        """
        Get a variable from its string in the scope of pc.
        """
        dvar = self._dvar_containers[var_name]
        return dvar.from_pc(pc_addr)

    def from_name(self, var_name: str) -> DebugVariableContainer:
        """
        Get the variable container for all variables named var_name

        :param var_name:    name for a variable
        """
        if var_name not in self._dvar_containers:
            self._dvar_containers[var_name] = DebugVariableContainer()
        return self._dvar_containers[var_name]

    def __getitem__(self, var_name: str) -> DebugVariableContainer:
        assert isinstance(var_name, str)
        return self.from_name(var_name)

    def add_variable(self, cle_var: Variable, low_pc: int, high_pc: int):
        """
        Add/load a variable

        :param cle_variable:    The variable to add
        :param low_pc:          Start of the visibility scope of the variable as program counter address (rebased)
        :param high_pc:         End of the visibility scope of the variable as program counter address (rebased)
        """
        name = cle_var.name
        assert name is not None
        if name not in self._dvar_containers:
            self._dvar_containers[name] = DebugVariableContainer()
        container = self._dvar_containers[name]
        container[low_pc:high_pc] = cle_var

    def __setitem__(self, index, cle_var):
        assert isinstance(index, slice) and isinstance(cle_var, Variable)
        return self.add_variable(cle_var, index.start, index.stop)

    # Methods similar to the once in VariableManager
    def add_variable_list(self, vlist: list[Variable], low_pc: int, high_pc: int):
        """
        Add all variables in a list with the same visibility range

        :param vlist:       A list of cle variables to add
        :param low_pc:      Start of the visibility scope as program counter address (rebased)
        :param high_pc:     End of the visibility scope as program counter address (rebased)
        """
        for v in vlist:
            self.add_variable(v, low_pc, high_pc)

    def load_from_dwarf(self, elf_object: ELF = None, cu: CompilationUnit = None):
        """
        Automatically load all variables (global/local) from the DWARF debugging info

        :param elf_object:  Optional, when only one elf object should be considered (e.g. p.loader.main_object)
        :param cu:          Optional, when only one compilation unit should be considered
        """
        objs = [elf_object] if elf_object else self._kb._project.loader.all_elf_objects
        for obj in objs:
            if cu:
                if obj not in obj.compilation_units:
                    break
                cu_list = [cu]
            else:
                cu_list = obj.compilation_units

            for cu_curr in cu_list:
                for cle_var in cu_curr.global_variables:
                    if cle_var.external:
                        self.add_variable(cle_var, obj.min_addr, obj.max_addr)
                    else:
                        # static variable
                        self.add_variable(cle_var, cu_curr.min_addr, cu_curr.max_addr)
                for subp in cu_curr.functions.values():
                    for cle_var in subp.local_variables:
                        low_pc = cle_var.lexical_block.low_pc
                        high_pc = cle_var.lexical_block.high_pc
                        self.add_variable(cle_var, low_pc, high_pc)


KnowledgeBasePlugin.register_default("dvars", DebugVariableManager)



"""
Spilling CFG Graph implementation with LRU caching and LMDB persistence.

This module provides SpillingCFGNodeDict and SpillingCFGGraph classes that implement
disk-backed storage for CFGNode instances, following the SpillingFunctionDict pattern.
"""

from __future__ import annotations

import logging
import threading
import weakref
import os
from collections import OrderedDict, defaultdict
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING, overload, Literal

import lmdb
import networkx
from archinfo.arch_soot import SootAddressDescriptor

from angr.protos import cfg_pb2
from .block_id import BlockID
from .cfg_node import CFGNode, CFGENode
from .spilling_digraph import SpillingDiGraph
from .types import CFGNODE_K, CFGENODE_K, SOOTNODE_K, K, CFG_ADDR_TYPES

if TYPE_CHECKING:
    from angr.knowledge_plugins.rtdb.rtdb import RuntimeDb
    from .cfg_model import CFGModel

l = logging.getLogger(name=__name__)


# a global flag to disable SpillingFunctionDict usage; mainly for testing purposes
USE_SPILLING_CFGNODE_DICT = os.environ.get("USE_SPILLING_CFGNODE_DICT", "True").lower() not in ("0", "false", "no")


class SpillingCFGNodeDict:
    """
    A dict-like container for CFGNode instances with LRU caching and LMDB spilling.

    This class keeps only the most recently accessed N nodes in memory, spilling others to an LMDB database on disk.

    :ivar cache_limit:          The maximum number of nodes to keep in memory.
    :ivar rtdb:                 A reference to the RuntimeDb knowledge base plugin.
    :ivar _lru_order:           An OrderedDict tracking the eviction order of cached nodes.
    :ivar _spilled_keys:        A set of block_ids that have been spilled to LMDB.
    :ivar _db_batch_size:       The number of nodes that are evicted in a single batch.
    :ivar _eviction_enabled:    A flag indicating whether eviction is currently enabled.
    """

    def __init__(
        self,
        rtdb: RuntimeDb | None,
        cfg_model: CFGModel | None = None,
        cache_limit: int = 1000,
        db_batch_size: int = 200,
    ):
        self._data: dict[K, CFGNode] = {}
        self._spilled_keys: set[K] = set()

        self._cache_limit: int = cache_limit
        self._db_batch_size: int = db_batch_size

        self.rtdb: RuntimeDb | None = rtdb
        self._cfg_model_ref: weakref.ref[CFGModel] | None = weakref.ref(cfg_model) if cfg_model is not None else None

        self._lru_order: OrderedDict[K, None] = OrderedDict()

        self._nodesdb = None
        self._eviction_enabled: bool = True
        self._loading_from_lmdb: bool = False
        self._db_load_lock = threading.Lock()
        self._db_store_lock = threading.Lock()

    def __del__(self):
        self._cleanup_lmdb()

    @property
    def _cfg_model(self) -> CFGModel | None:
        if self._cfg_model_ref is None:
            return None
        return self._cfg_model_ref()

    @_cfg_model.setter
    def _cfg_model(self, value: CFGModel | None) -> None:
        self._cfg_model_ref = weakref.ref(value) if value is not None else None

    def __getitem__(self, block_key: K) -> CFGNode:
        # First try to get from in-memory cache
        if block_key in self._data:
            self._touch(block_key)
            return self._data[block_key]

        # Try to load from LMDB if spilled
        if block_key in self._spilled_keys:
            node = self._load_from_lmdb(block_key)
            if node is not None:
                return node

        raise KeyError(block_key)

    def __setitem__(self, block_key: K, node: CFGNode) -> None:
        self._data[block_key] = node
        self._on_node_stored(block_key)

    def __delitem__(self, block_key: K) -> None:
        # Remove from in-memory cache if present
        if block_key in self._data:
            del self._data[block_key]
        # Remove from spilled set if present
        self._spilled_keys.discard(block_key)
        # Remove from LRU order
        if block_key in self._lru_order:
            del self._lru_order[block_key]

    def __contains__(self, block_key: K) -> bool:
        return block_key in self._data or block_key in self._spilled_keys

    def __len__(self) -> int:
        return len(self._data) + len(self._spilled_keys)

    def __iter__(self) -> Iterator[K]:
        # Iterate over all block_keys
        yield from list(self._data) + list(self._spilled_keys)

    def get(self, block_key: K, default: CFGNode | None = None) -> CFGNode | None:
        try:
            return self[block_key]
        except KeyError:
            return default

    def keys(self) -> Iterator[K]:
        return iter(self)

    def values(self) -> Iterator[CFGNode]:
        for block_key in self:
            yield self[block_key]

    def items(self) -> Iterator[tuple[K, CFGNode]]:
        for block_key in self:
            yield block_key, self[block_key]

    def clear(self) -> None:
        self._data.clear()
        self._lru_order.clear()
        self._spilled_keys.clear()
        self._cleanup_lmdb()

    def copy(self) -> SpillingCFGNodeDict:
        new_dict = SpillingCFGNodeDict(
            self.rtdb,
            self._cfg_model,
            cache_limit=self._cache_limit,
            db_batch_size=self._db_batch_size,
        )
        # Temporarily disable eviction during copy
        new_dict._eviction_enabled = False

        # Copy in-memory nodes
        for block_key, node in self._data.items():
            new_dict._data[block_key] = node.copy()
            new_dict._lru_order[block_key] = None

        # Copy spilled data from LMDB
        if self._spilled_keys and self._nodesdb is not None and self.rtdb is not None:
            new_dict._init_lmdb()
            assert new_dict._nodesdb is not None
            with (
                self.rtdb.begin_txn(self._nodesdb) as src_txn,
                self.rtdb.begin_txn(new_dict._nodesdb, write=True) as dst_txn,
            ):
                for block_key in self._spilled_keys:
                    key = str(block_key).encode("utf-8")
                    value = src_txn.get(key)
                    if value is not None:
                        dst_txn.put(key, value)
                        new_dict._spilled_keys.add(block_key)

        new_dict._eviction_enabled = True
        return new_dict

    #
    # Properties
    #

    @property
    def cache_limit(self) -> int:
        return self._cache_limit

    @cache_limit.setter
    def cache_limit(self, value: int) -> None:
        self._cache_limit = value
        # Trigger eviction if we're over the new limit
        if self.cached_count > value + self._db_batch_size:
            self._evict_lru()

    @property
    def db_batch_size(self) -> int:
        return self._db_batch_size

    @db_batch_size.setter
    def db_batch_size(self, value: int) -> None:
        self._db_batch_size = value
        if self.cached_count > self._cache_limit + value:
            self._evict_lru()

    @property
    def cached_count(self) -> int:
        return len(self._data)

    @property
    def spilled_count(self) -> int:
        return len(self._spilled_keys)

    @property
    def total_count(self) -> int:
        return len(self._spilled_keys) + len(self._data)

    def is_cached(self, block_key: K) -> bool:
        return block_key in self._data

    #
    # LRU Cache Management
    #

    def _touch(self, block_key: K) -> None:
        if block_key in self._lru_order:
            self._lru_order.move_to_end(block_key)
        else:
            self._lru_order[block_key] = None

    def _on_node_stored(self, block_key: K) -> None:
        self._touch(block_key)
        self._spilled_keys.discard(block_key)

        if (
            self._eviction_enabled
            and self._cache_limit is not None
            and self.cached_count > self._cache_limit + self._db_batch_size
        ):
            self._evict_lru()

    def _evict_lru(self) -> bool:
        with self._db_store_lock:
            evicted_any = False
            while self.cached_count > self._cache_limit + self._db_batch_size:
                # Evict enough to get below the limit, in batches
                to_evict = self.cached_count - self._cache_limit
                batch_size = min(self._db_batch_size, to_evict)
                if self._evict_n(batch_size) == 0:
                    break
                evicted_any = True
            return evicted_any

    def _evict_n(self, n: int) -> int:
        if not self._lru_order:
            return 0

        evicted = 0
        nodes_to_save = []
        nodes_to_evict = []
        for lru_block_key in self._lru_order:
            if evicted >= n:
                break

            if lru_block_key not in self._data:
                nodes_to_evict.append(lru_block_key)
                continue

            node = self._data[lru_block_key]
            if node.dirty:
                nodes_to_save.append((lru_block_key, node))

            del self._data[lru_block_key]
            nodes_to_evict.append(lru_block_key)
            self._spilled_keys.add(lru_block_key)
            evicted += 1

        for key in nodes_to_evict:
            del self._lru_order[key]

        if nodes_to_save:
            self._save_to_lmdb(nodes_to_save)

        return evicted

    #
    # LMDB Management
    #

    def _init_lmdb(self) -> None:
        if self._nodesdb is None and self.rtdb is not None:
            self._nodesdb = self.rtdb.open_db("cfgnodes")
            l.debug("Initialized CFGNode LMDB cache.")

    def _cleanup_lmdb(self) -> None:
        if self._nodesdb is not None and self.rtdb is not None:
            self.rtdb.drop_db(self._nodesdb)
            self._nodesdb = None

    def _save_to_lmdb(self, nodes: list[tuple[K, CFGNode]]) -> None:
        if self.rtdb is None:
            return

        self._init_lmdb()
        assert self._nodesdb is not None

        while True:
            try:
                with self.rtdb.begin_txn(self._nodesdb, write=True) as txn:
                    for block_key, node in nodes:
                        cmsg = node.serialize_to_cmessage()
                        payload = cmsg.SerializeToString()
                        # Prefix with type byte: 0x00 for CFGNode, 0x01 for CFGENode
                        data = b"\x01" + payload if isinstance(node, CFGENode) else b"\x00" + payload
                        key = str(block_key).encode("utf-8")
                        txn.put(key, data)
                break
            except lmdb.MapFullError:
                self.rtdb.increase_lmdb_map_size()

    def _load_from_lmdb(self, block_key: K) -> CFGNode | None:
        if self._nodesdb is None or self.rtdb is None:
            return None

        with self._db_load_lock:
            return self._load_from_lmdb_core(block_key)

    def _load_from_lmdb_core(self, block_key: K) -> CFGNode | None:
        if self._loading_from_lmdb:
            raise RuntimeError("Recursive loading from LMDB detected. This is a bug.")

        assert self.rtdb is not None and self._nodesdb is not None

        self._loading_from_lmdb = True

        try:
            key = str(block_key).encode("utf-8")

            with self.rtdb.begin_txn(self._nodesdb) as txn:
                value = txn.get(key)
                if value is None:
                    return None

                type_byte = value[0]
                payload = value[1:]

                if type_byte == 0x00:
                    cmsg = cfg_pb2.CFGNode()  # type:ignore
                    cmsg.ParseFromString(payload)
                    node = CFGNode.parse_from_cmessage(cmsg, cfg=self._cfg_model)
                elif type_byte == 0x01:
                    cmsg = cfg_pb2.CFGENode()  # type:ignore
                    cmsg.ParseFromString(payload)
                    node = CFGENode.parse_from_cmessage(cmsg, cfg=self._cfg_model)
                else:
                    raise ValueError(f"Unknown node type byte: {type_byte:#x}")

                node.dirty = False

            # Remove from spilled set and add to cache
            self._spilled_keys.discard(block_key)
            self._data[block_key] = node
            self._on_node_stored(block_key)

            return node
        finally:
            self._loading_from_lmdb = False

            if (
                self._eviction_enabled
                and self._cache_limit is not None
                and self.cached_count > self._cache_limit + self._db_batch_size
            ):
                self._evict_lru()

    def load_all_spilled(self) -> None:
        if not self._spilled_keys:
            return

        old_eviction_state = self._eviction_enabled
        self._eviction_enabled = False

        try:
            block_keys_to_load = list(self._spilled_keys)
            for block_key in block_keys_to_load:
                self._load_from_lmdb(block_key)
        finally:
            self._eviction_enabled = old_eviction_state

    def evict_all_cached(self) -> None:
        if self.cached_count == 0:
            return
        with self._db_store_lock:
            self._evict_n(self.cached_count)

    #
    # Pickling
    #

    def __getstate__(self):
        # Load all spilled nodes before pickling
        self.load_all_spilled()
        return {
            "cache_limit": self._cache_limit,
            "db_batch_size": self._db_batch_size,
            "items": dict(self._data),
        }

    def __setstate__(self, state: dict):
        self._cache_limit = state["cache_limit"]
        self._db_batch_size = state["db_batch_size"]
        self._data = {}
        self.rtdb = None
        self._cfg_model_ref = None
        self._lru_order = OrderedDict()
        self._spilled_keys = set()
        self._nodesdb = None
        self._eviction_enabled = True
        self._loading_from_lmdb = False
        self._db_load_lock = threading.Lock()
        self._db_store_lock = threading.Lock()

        for k, v in state["items"].items():
            self[k] = v


class _AdjacencyDict:
    """Helper class to support graph[src][dst] access pattern."""

    def __init__(self, graph: SpillingCFG, src_block_key: K):
        self._graph = graph
        self._src_block_key = src_block_key

    def __getitem__(self, dst_node: CFGNode) -> dict:
        dst_block_key = get_block_key(dst_node)
        return self._graph._graph[self._src_block_key][dst_block_key]

    def __contains__(self, dst_node: CFGNode) -> bool:
        dst_block_key = get_block_key(dst_node)
        return dst_block_key in self._graph._graph[self._src_block_key]

    def keys(self) -> Iterator[CFGNode]:
        for dst_block_key in self._graph._graph[self._src_block_key]:
            yield self._graph.get_node_by_key(dst_block_key)

    def values(self) -> Iterator[dict]:
        for dst_block_key in self._graph._graph[self._src_block_key]:
            yield self._graph._graph[self._src_block_key][dst_block_key]

    def items(self) -> Iterator[tuple[CFGNode, dict]]:
        for dst_block_key in self._graph._graph[self._src_block_key]:
            yield self._graph.get_node_by_key(dst_block_key), self._graph._graph[self._src_block_key][dst_block_key]

    def __iter__(self) -> Iterator[CFGNode]:
        return self.keys()


class _NodeView:
    """View over graph nodes supporting len(), iteration, and call with data=True."""

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    def __len__(self) -> int:
        return len(self._graph._graph)

    def __iter__(self) -> Iterator[CFGNode]:
        for block_key in self._graph._graph.nodes():
            yield self._graph.get_node_by_key(block_key)

    @overload
    def __call__(self, data: Literal[False] = False) -> Iterator[CFGNode]: ...
    @overload
    def __call__(self, data: Literal[True]) -> Iterator[tuple[CFGNode, dict]]: ...

    def __call__(self, data: bool = False) -> Iterator[CFGNode] | Iterator[tuple[CFGNode, dict]]:
        if data:
            for block_key, node_data in self._graph._graph.nodes(data=True):
                yield self._graph.get_node_by_key(block_key), node_data
        else:
            yield from self

    def __contains__(self, node: CFGNode) -> bool:
        return self._graph.has_node(node)


class _EdgeView:
    """View over graph edges supporting len(), iteration, and call with data=True."""

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    def __len__(self) -> int:
        return self._graph._graph.number_of_edges()

    def __iter__(self) -> Iterator[tuple[CFGNode, CFGNode]]:
        for src_id, dst_id in self._graph._graph.edges():
            yield self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id)

    @overload
    def __call__(self, data: Literal[False] = False) -> Iterator[tuple[CFGNode, CFGNode]]: ...
    @overload
    def __call__(self, data: Literal[True]) -> Iterator[tuple[CFGNode, CFGNode, dict]]: ...

    def __call__(
        self, data: bool = False
    ) -> Iterator[tuple[CFGNode, CFGNode]] | Iterator[tuple[CFGNode, CFGNode, dict]]:
        if data:
            for src_id, dst_id, edge_data in self._graph._graph.edges(data=True):
                yield self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id), edge_data
        else:
            yield from self


class _InEdgeView:
    """View over graph in-edges supporting call, subscript, len, and iteration.

    Supports both ``graph.in_edges(node)`` and ``graph.in_edges[node]``.
    """

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    @overload
    def __call__(self, nbunch=None, data: Literal[False] = False) -> list[tuple[CFGNode, CFGNode]]: ...
    @overload
    def __call__(self, nbunch=None, *, data: Literal[True]) -> list[tuple[CFGNode, CFGNode, dict]]: ...

    def __call__(
        self, nbunch=None, data: bool = False
    ) -> list[tuple[CFGNode, CFGNode]] | list[tuple[CFGNode, CFGNode, dict]]:
        if nbunch is not None:
            nbunch = [get_block_key(nbunch)] if isinstance(nbunch, CFGNode) else [get_block_key(n) for n in nbunch]

        if data:
            return [
                (self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id), edge_data)
                for src_id, dst_id, edge_data in self._graph._graph.in_edges(nbunch, data=True)
            ]
        return [
            (self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id))
            for src_id, dst_id in self._graph._graph.in_edges(nbunch)
        ]

    def __getitem__(self, node) -> list[tuple[CFGNode, CFGNode]]:
        return self(node)

    def __iter__(self) -> Iterator[tuple[CFGNode, CFGNode]]:
        for src_id, dst_id in self._graph._graph.in_edges():
            yield self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id)

    def __len__(self) -> int:
        return self._graph._graph.number_of_edges()


class _OutEdgeView:
    """View over graph out-edges supporting call, subscript, len, and iteration.

    Supports both ``graph.out_edges(node)`` and ``graph.out_edges[node]``.
    """

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    @overload
    def __call__(self, nbunch=None, data: Literal[False] = False) -> list[tuple[CFGNode, CFGNode]]: ...
    @overload
    def __call__(self, nbunch=None, *, data: Literal[True]) -> list[tuple[CFGNode, CFGNode, dict]]: ...

    def __call__(
        self, nbunch=None, data: bool = False
    ) -> list[tuple[CFGNode, CFGNode]] | list[tuple[CFGNode, CFGNode, dict]]:
        if nbunch is not None:
            nbunch = [get_block_key(nbunch)] if isinstance(nbunch, CFGNode) else [get_block_key(n) for n in nbunch]

        if data:
            return [
                (self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id), edge_data)
                for src_id, dst_id, edge_data in self._graph._graph.out_edges(nbunch, data=True)
            ]
        return [
            (self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id))
            for src_id, dst_id in self._graph._graph.out_edges(nbunch)
        ]

    def __getitem__(self, node) -> list[tuple[CFGNode, CFGNode]]:
        return self(node)

    def __iter__(self) -> Iterator[tuple[CFGNode, CFGNode]]:
        for src_id, dst_id in self._graph._graph.out_edges():
            yield self._graph.get_node_by_key(src_id), self._graph.get_node_by_key(dst_id)

    def __len__(self) -> int:
        return self._graph._graph.number_of_edges()


class _InDegreeView:
    """View over graph in-degrees supporting call, subscript, len, and iteration.

    Supports both ``graph.in_degree(node)`` and ``graph.in_degree[node]``.
    """

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    def __call__(self, node: CFGNode | None = None) -> int | Iterator[tuple[CFGNode, int]]:
        if node is None:
            return iter(self)
        return self[node]

    def __getitem__(self, node: CFGNode) -> int:
        block_key = get_block_key(node)
        if block_key not in self._graph._graph:
            return 0
        return self._graph._graph.in_degree(block_key)

    def __iter__(self) -> Iterator[tuple[CFGNode, int]]:
        for block_key, deg in self._graph._graph.in_degree():
            yield self._graph.get_node_by_key(block_key), deg

    def __len__(self) -> int:
        return len(self._graph._graph)


class _OutDegreeView:
    """View over graph out-degrees supporting call, subscript, len, and iteration.

    Supports both ``graph.out_degree(node)`` and ``graph.out_degree[node]``.
    """

    def __init__(self, graph: SpillingCFG):
        self._graph = graph

    def __call__(self, node: CFGNode | None = None) -> int | Iterator[tuple[CFGNode, int]]:
        if node is None:
            return iter(self)
        return self[node]

    def __getitem__(self, node: CFGNode) -> int:
        block_key = get_block_key(node)
        return self._graph._out_degree_cache.get(block_key, 0)

    def __iter__(self) -> Iterator[tuple[CFGNode, int]]:
        for block_key, deg in self._graph._out_degree_cache.items():
            yield self._graph.get_node_by_key(block_key), deg

    def __len__(self) -> int:
        return len(self._graph._out_degree_cache)


@overload
def get_block_key(node: CFGNode) -> CFGNODE_K | SOOTNODE_K: ...
@overload
def get_block_key(node: CFGENode) -> CFGENODE_K: ...  # type:ignore


def get_block_key(node: CFGNode | CFGENode) -> K:
    """
    Get the unique identifier for a CFGNode. Typically this unique identifier contains the address of the block and
    the looping_times of the block (in case there are multiple blocks with the same address, which may happen after
    loop unrolling in a CFGEmulated instance).

    :param node:    The CFGNode or CFGENode instance to get the block key for.
    :return:        The unique identifier.
    """

    if isinstance(node.addr, SootAddressDescriptor):
        return node.addr

    block_id = node.block_id
    if block_id is None:
        block_id = node.addr
    block_key = block_id, node.size if node.size is not None else -1

    if isinstance(node, CFGENode):
        return *block_key, node.looping_times if node.looping_times is not None else 0
    return block_key


@overload
def block_key_to_addr(block_key: CFGNODE_K | CFGENODE_K) -> int: ...
@overload
def block_key_to_addr(block_key: SOOTNODE_K) -> SootAddressDescriptor: ...


def block_key_to_addr(block_key: K) -> int | SootAddressDescriptor:
    """Extract the address from a block key."""
    if isinstance(block_key, SootAddressDescriptor):
        return block_key
    if isinstance(block_key, tuple) and len(block_key) >= 2:
        item = block_key[0]
        if isinstance(item, BlockID):
            return item.addr
        assert isinstance(item, int)
        return item
    raise ValueError(f"Invalid block key format: {block_key!r}")


def block_key_to_size(block_key: K) -> int | None:
    """Extract the size from a block key, if present."""
    if isinstance(block_key, SootAddressDescriptor):
        return None
    if isinstance(block_key, tuple) and len(block_key) >= 2:
        size = block_key[1]
        return size if size != -1 else None
    raise ValueError(f"Invalid block key format: {block_key!r}")


class SpillingCFG:
    """
    A graph wrapper that stores CFGNode instances in a spilling dict while keeping only primitive keys in the
    underlying networkx graph.

    This provides a networkx-compatible interface while supporting disk-backed storage for large CFGs.

    addr_type must be "int", "block_id", or "soot". You can change addr_type before the first node is inserted but not
    after, since it affects how keys are serialized and deserialized.
    """

    _addr_type: CFG_ADDR_TYPES

    def __init__(
        self,
        rtdb: RuntimeDb | None = None,
        cfg_model: CFGModel | None = None,
        cache_limit: int | None = None,
        db_batch_size: int = 800,
        edge_cache_limit: int | None = None,
        edge_db_batch_size: int = 800,
        addr_type: CFG_ADDR_TYPES = "int",
    ):
        if USE_SPILLING_CFGNODE_DICT:
            effective_edge_cache_limit = edge_cache_limit if edge_cache_limit is not None else 2**31 - 1
        else:
            effective_edge_cache_limit = 2**31 - 1

        self._addr_type = addr_type
        self._graph: SpillingDiGraph = SpillingDiGraph(
            rtdb=rtdb,
            edge_cache_limit=effective_edge_cache_limit,
            db_batch_size=edge_db_batch_size,
            addr_type=addr_type,
        )
        self._cfg_model_ref: weakref.ref[CFGModel] | None = weakref.ref(cfg_model) if cfg_model is not None else None
        self._rtdb = rtdb

        if USE_SPILLING_CFGNODE_DICT:
            effective_cache_limit = cache_limit if cache_limit is not None else 2**31 - 1
        else:
            effective_cache_limit = 2**31 - 1

        self._nodes: SpillingCFGNodeDict = SpillingCFGNodeDict(
            rtdb,
            cfg_model,
            cache_limit=effective_cache_limit,
            db_batch_size=db_batch_size,
        )
        self._keys_by_addr: dict[int, set[K]] = defaultdict(set)
        self._call_dst_keys: set[K] = set()
        self._out_degree_cache: dict[K, int] = {}
        self._spilling_enabled = cache_limit is not None
        self._edge_spilling_enabled = edge_cache_limit is not None

    @property
    def addr_type(self) -> str:
        return self._addr_type

    @addr_type.setter
    def addr_type(self, value: str) -> None:
        if value not in ("int", "block_id", "soot"):
            raise ValueError("addr_type must be 'int', 'block_id', or 'soot'")
        if self._nodes.total_count > 0:
            raise RuntimeError("Cannot change addr_type after nodes have been added")
        self._addr_type = value
        self._graph.addr_type = value

    @property
    def _cfg_model(self) -> CFGModel | None:
        if self._cfg_model_ref is None:
            return None
        return self._cfg_model_ref()

    @_cfg_model.setter
    def _cfg_model(self, value: CFGModel | None) -> None:
        self._cfg_model_ref = weakref.ref(value) if value is not None else None
        self._nodes._cfg_model = value

    def get_node_by_key(self, block_key: K) -> CFGNode:
        """Get a CFGNode by block_id, with fallback to graph node data."""
        # First try the nodes dict (handles spilling)
        if block_key in self._nodes:
            return self._nodes[block_key]
        raise KeyError(block_key)

    @property
    def node_keys(self) -> Iterator[K]:
        """Get an iterator over all block_keys in the graph."""
        yield from self._nodes

    #
    # Node operations
    #

    def add_node(self, node: CFGNode, **attr) -> None:
        block_key = get_block_key(node)
        self._nodes[block_key] = node
        self._graph.add_node(block_key, **attr)
        # update _keys_by_addr
        self._keys_by_addr[node.addr].add(block_key)

    def remove_node(self, node: CFGNode) -> None:
        block_key = get_block_key(node)
        if block_key in self._nodes:
            del self._nodes[block_key]
        self._keys_by_addr[node.addr].discard(block_key)
        if not self._keys_by_addr.get(node.addr):
            self._keys_by_addr.pop(node.addr, None)
        if block_key in self._graph:
            # Update call destination cache: remove this node as a destination
            self._call_dst_keys.discard(block_key)
            # For each outgoing call edge, check if the successor loses all incoming call edges
            for _, succ_key, edata in self._graph.out_edges(block_key, data=True):
                ejk = edata.get("jumpkind", "")
                if (ejk == "Ijk_Call" or ejk.startswith("Ijk_Sys")) and succ_key in self._call_dst_keys:
                    has_other_call = False
                    for pred_key, _, pred_edata in self._graph.in_edges(succ_key, data=True):
                        if pred_key == block_key:
                            continue
                        pjk = pred_edata.get("jumpkind", "")
                        if pjk == "Ijk_Call" or pjk.startswith("Ijk_Sys"):
                            has_other_call = True
                            break
                    if not has_other_call:
                        self._call_dst_keys.discard(succ_key)

            # Decrement out_degree_cache for each predecessor
            for pred_key in self._graph.predecessors(block_key):
                if pred_key in self._out_degree_cache:
                    self._out_degree_cache[pred_key] -= 1
            self._graph.remove_node(block_key)
        self._out_degree_cache.pop(block_key, None)

    def has_node(self, node: CFGNode) -> bool:
        block_key = get_block_key(node)
        return block_key in self._graph

    def nodes_by_addr(self, addr: int) -> Iterator[CFGNode]:
        for block_key in self._keys_by_addr.get(addr, []):
            yield self.get_node_by_key(block_key)

    def has_node_addr(self, addr: int) -> bool:
        return addr in self._keys_by_addr

    def __contains__(self, node: CFGNode) -> bool:
        return self.has_node(node)

    def __len__(self) -> int:
        return len(self._graph)

    def number_of_nodes(self) -> int:
        return len(self._graph)

    @property
    def nodes(self) -> _NodeView:
        """Return a view of nodes supporting len(), iteration, and call with data=True."""
        return _NodeView(self)

    def __iter__(self) -> Iterator[CFGNode]:
        for block_key in self._graph:
            yield self.get_node_by_key(block_key)

    #
    # Edge operations
    #

    def add_edge(self, src: CFGNode, dst: CFGNode, **attr) -> None:
        src_block_key = get_block_key(src)
        dst_block_key = get_block_key(dst)

        # Always update _nodes with the passed nodes
        # This is needed for node replacement during _shrink_node
        self._nodes[src_block_key] = src
        self._nodes[dst_block_key] = dst

        self.add_edge_by_key(src_block_key, dst_block_key, **attr)

    def add_edge_by_key(self, src_block_key: K, dst_block_key: K, **attr) -> None:
        # Ensure nodes exist in the graph structure
        if src_block_key not in self._graph:
            self._graph.add_node(src_block_key)

        if dst_block_key not in self._graph:
            self._graph.add_node(dst_block_key)

        # Update out_degree_cache before adding edge (only increment if edge is new)
        has_edge = self._graph.has_edge(src_block_key, dst_block_key)
        self._graph.add_edge(src_block_key, dst_block_key, **attr)
        if not has_edge:
            self._out_degree_cache[src_block_key] = self._out_degree_cache.get(src_block_key, 0) + 1

        # Track call destination keys
        jumpkind = attr.get("jumpkind", "")
        if jumpkind == "Ijk_Call" or jumpkind.startswith("Ijk_Sys"):
            self._call_dst_keys.add(dst_block_key)

    def remove_edge(self, src: CFGNode, dst: CFGNode) -> None:
        src_block_key = get_block_key(src)
        dst_block_key = get_block_key(dst)

        # Update call destination cache before removing the edge
        if dst_block_key in self._call_dst_keys:
            edge_data = self._graph.get_edge_data(src_block_key, dst_block_key)
            if edge_data is not None:
                jk = edge_data.get("jumpkind", "")
                if jk == "Ijk_Call" or jk.startswith("Ijk_Sys"):
                    # Check if dst has any other incoming call edges
                    has_other_call = False
                    for pred_key, _, edata in self._graph.in_edges(dst_block_key, data=True):
                        if pred_key == src_block_key:
                            continue
                        ejk = edata.get("jumpkind", "")
                        if ejk == "Ijk_Call" or ejk.startswith("Ijk_Sys"):
                            has_other_call = True
                            break
                    if not has_other_call:
                        self._call_dst_keys.discard(dst_block_key)

        self._graph.remove_edge(src_block_key, dst_block_key)
        if src_block_key in self._out_degree_cache:
            self._out_degree_cache[src_block_key] -= 1

    def has_edge(self, src: CFGNode, dst: CFGNode) -> bool:
        src_block_key = get_block_key(src)
        dst_block_key = get_block_key(dst)
        return self._graph.has_edge(src_block_key, dst_block_key)

    def get_edge_data(self, src: CFGNode, dst: CFGNode, default=None) -> dict | None:
        src_block_key = get_block_key(src)
        dst_block_key = get_block_key(dst)
        return self._graph.get_edge_data(src_block_key, dst_block_key, default)

    def number_of_edges(self) -> int:
        return self._graph.number_of_edges()

    @property
    def edges(self) -> _EdgeView:
        """Return a view of edges supporting len(), iteration, and call with data=True."""
        return _EdgeView(self)

    #
    # Call destination cache
    #

    @property
    def call_destination_keys(self) -> set[K]:
        """Return the set of block keys that are destinations of call/syscall edges."""
        return self._call_dst_keys

    def call_destination_nodes(self) -> Iterator[CFGNode]:
        """Yield CFGNode for each call/syscall destination."""
        for key in self._call_dst_keys:
            yield self.get_node_by_key(key)

    #
    # Neighbor operations
    #

    def predecessors(self, node: CFGNode) -> Iterator[CFGNode]:
        block_key = get_block_key(node)
        for pred_key in self._graph.predecessors(block_key):
            yield self.get_node_by_key(pred_key)

    def successors(self, node: CFGNode) -> Iterator[CFGNode]:
        block_key = get_block_key(node)
        for succ_key in self._graph.successors(block_key):
            yield self.get_node_by_key(succ_key)

    @property
    def in_edges(self) -> _InEdgeView:
        """Return a view of in-edges supporting call, subscript, len, and iteration."""
        return _InEdgeView(self)

    @property
    def out_edges(self) -> _OutEdgeView:
        """Return a view of out-edges supporting call, subscript, len, and iteration."""
        return _OutEdgeView(self)

    @property
    def in_degree(self) -> _InDegreeView:
        """Return a view of in-degrees supporting call, subscript, len, and iteration."""
        return _InDegreeView(self)

    @property
    def out_degree(self) -> _OutDegreeView:
        """Return a view of out-degrees supporting call, subscript, len, and iteration."""
        return _OutDegreeView(self)

    @overload
    def out_edges_by_key(self, key: K, data: Literal[False] = False) -> Generator[tuple[K, K]]: ...
    @overload
    def out_edges_by_key(self, key: K, *, data: Literal[True]) -> Generator[tuple[K, K, dict]]: ...

    def out_edges_by_key(self, key: K, data: bool = False) -> Generator[tuple[K, K] | tuple[K, K, dict]]:
        if key not in self._graph:
            return
        yield from self._graph.out_edges(key, data=data)

    def out_degree_by_key(self, key: K) -> int:
        return self._out_degree_cache.get(key, 0)

    #
    # Adjacency access
    #

    def __getitem__(self, node: CFGNode) -> _AdjacencyDict:
        block_key = get_block_key(node)
        if block_key not in self._graph:
            raise KeyError(node)
        return _AdjacencyDict(self, block_key)

    #
    # Graph operations
    #

    def copy(self) -> SpillingCFG:
        new_graph = SpillingCFG(
            rtdb=self._rtdb,
            cfg_model=self._cfg_model,
            cache_limit=self._nodes._cache_limit if self._spilling_enabled else None,
            db_batch_size=self._nodes.db_batch_size,
            edge_cache_limit=self._graph._edge_cache_limit if self._edge_spilling_enabled else None,
            edge_db_batch_size=self._graph._edge_db_batch_size,
            addr_type=self._addr_type,
        )

        new_graph._nodes = self._nodes.copy()
        new_graph._spilling_enabled = self._spilling_enabled
        new_graph._edge_spilling_enabled = self._edge_spilling_enabled
        new_graph._keys_by_addr = defaultdict(set)
        for addr, keys in self._keys_by_addr.items():
            new_graph._keys_by_addr[addr] = set(keys)
        new_graph._call_dst_keys = set(self._call_dst_keys)
        new_graph._graph = self._graph.copy()
        new_graph._out_degree_cache = dict(self._out_degree_cache)

        return new_graph

    def subgraph(self, nodes) -> networkx.DiGraph:
        """
        Return a subgraph as a regular networkx DiGraph with CFGNode instances.
        This is useful for algorithms that need a pure networkx graph.
        """
        block_keys = [get_block_key(n) for n in nodes]
        sub = self._graph.subgraph(block_keys)

        # Convert to CFGNode-based graph
        result = networkx.DiGraph()
        for block_key in sub.nodes():
            result.add_node(self.get_node_by_key(block_key))
        for src_key, dst_key, data in sub.edges(data=True):
            result.add_edge(self.get_node_by_key(src_key), self.get_node_by_key(dst_key), **data)

        return result

    def to_networkx(self) -> networkx.DiGraph:
        """
        Convert to a pure networkx DiGraph with CFGNode instances as nodes.
        Warning: This loads all spilled nodes into memory.
        """
        result = networkx.DiGraph()
        for node in self.nodes():
            result.add_node(node)
        for src, dst, data in self.edges(data=True):
            result.add_edge(src, dst, **data)
        return result

    def from_networkx(self, nx_graph: networkx.DiGraph) -> None:
        """
        Load graph structure from a networkx DiGraph with CFGNode instances as nodes.
        """
        self._out_degree_cache.clear()
        self._call_dst_keys.clear()
        self._graph.clear()
        self._nodes.clear()

        for node in nx_graph.nodes():
            self.add_node(node)

        for src, dst, data in nx_graph.edges(data=True):
            self.add_edge(src, dst, **data)

    #
    # Spilling control
    #

    @property
    def cache_limit(self) -> int | None:
        if self._spilling_enabled:
            return self._nodes.cache_limit
        return None

    @cache_limit.setter
    def cache_limit(self, value: int | None) -> None:
        if value is not None:
            self._nodes.cache_limit = value
            self._spilling_enabled = True
        else:
            # Set to a very large value to effectively disable spilling
            self._nodes.cache_limit = 2**31 - 1
            self._spilling_enabled = False

    @property
    def db_batch_size(self) -> int:
        return self._nodes.db_batch_size

    @db_batch_size.setter
    def db_batch_size(self, value: int) -> None:
        self._nodes.db_batch_size = value

    @property
    def cached_count(self) -> int:
        return self._nodes.cached_count

    @property
    def spilled_count(self) -> int:
        return self._nodes.spilled_count

    def load_all_spilled(self) -> None:
        self._nodes.load_all_spilled()

    def evict_all_cached(self) -> None:
        self._nodes.evict_all_cached()

    #
    # Pickling
    #

    def __getstate__(self):
        self._nodes.load_all_spilled()
        self._graph.load_all_spilled_edges()
        nodes_state = self._nodes.__getstate__()

        return {
            "graph": self._graph,
            "nodes": nodes_state,
            "spilling_enabled": self._spilling_enabled,
            "edge_spilling_enabled": self._edge_spilling_enabled,
            "db_batch_size": self._nodes.db_batch_size,
            "addr_type": self._addr_type,
        }

    def __setstate__(self, state: dict):
        self._addr_type = state.get("addr_type", "int")
        self._graph = state["graph"]
        self._spilling_enabled = state["spilling_enabled"]
        self._edge_spilling_enabled = state.get("edge_spilling_enabled", False)
        self._cfg_model_ref = None
        self._rtdb = None
        self._keys_by_addr = defaultdict(set)

        nodes_state = state["nodes"]
        self._nodes = SpillingCFGNodeDict.__new__(SpillingCFGNodeDict)
        self._nodes.__setstate__(nodes_state)

        # initialize _keys_by_addr
        for node_key, node in self._nodes.items():
            self._keys_by_addr[node.addr].add(node_key)

        # rebuild _out_degree_cache from the graph
        self._out_degree_cache = {}
        for key, deg in self._graph.out_degree():
            self._out_degree_cache[key] = deg

        # rebuild call destination keys cache from edges
        self._call_dst_keys = set()
        for _, dst_key, edata in self._graph.edges(data=True):
            jk = edata.get("jumpkind", "")
            if jk == "Ijk_Call" or jk.startswith("Ijk_Sys"):
                self._call_dst_keys.add(dst_key)



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations
from angr.ailment.expression import Convert, Extract

from .base import PeepholeOptimizationExprBase


class RemoveNoopConversions(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Remove no-op conversions"
    expr_classes = (Convert, Extract)

    def optimize(self, expr: Convert | Extract, **kwargs):
        if isinstance(expr, Convert):
            inner = expr.operand
            signed = expr.is_signed
            ints = expr.from_type == expr.to_type == Convert.TYPE_INT
        else:
            inner = expr.base
            signed = False
            ints = True

        if inner.bits == expr.bits and ints:
            return inner

        if isinstance(inner, Convert):
            if (
                expr.bits < inner.bits
                and expr.bits == inner.operand.bits
                and ints
                and inner.from_type == inner.to_type == Convert.TYPE_INT
            ):
                # extension then truncation (e.g., 1->64->1) can be removed, but truncation then extension cannot be
                # removed (e.g., the high 32 bits must be removed during 64->32->64)
                return inner.operand
            if (
                expr.bits < inner.from_bits
                and inner.to_bits >= inner.from_bits
                and ints
                and inner.from_type == inner.to_type == Convert.TYPE_INT
            ):
                # merge extend and then bigtruncate into just truncate
                return Convert(
                    expr.idx,
                    inner.from_bits,
                    expr.bits,
                    signed,
                    inner.operand,
                    **expr.tags,
                )

        return None



# pylint:disable=missing-class-docstring
from __future__ import annotations

from pydantic import BaseModel


class VariableRename(BaseModel):
    old_name: str
    new_name: str


class VariableNameSuggestions(BaseModel):
    renames: list[VariableRename]


class FunctionNameSuggestion(BaseModel):
    function_name: str


class VariableTypeChange(BaseModel):
    variable_name: str
    new_type: str


class VariableTypeSuggestions(BaseModel):
    type_changes: list[VariableTypeChange]



from __future__ import annotations
from collections import defaultdict

from angr.ailment import AILBlockViewer
from angr.ailment.expression import Tmp
from angr.ailment.statement import Statement, Assignment
from angr.ailment.block import Block


class TmpUsesCollector(AILBlockViewer):
    """
    Collect all uses of temporary variables and their use statement IDs in an AIL block.
    """

    def __init__(self):
        super().__init__()

        self.tmp_and_uselocs: dict[tuple[int, int], set[tuple[Tmp, int]]] = defaultdict(set)

    def _handle_Tmp(self, expr_idx: int, expr: Tmp, stmt_idx: int, stmt: Statement, block: Block | None):
        if isinstance(stmt, Assignment) and expr is stmt.dst:
            return
        self.tmp_and_uselocs[(expr.tmp_idx, expr.bits)].add((expr, stmt_idx))



from __future__ import annotations
import angr.ailment as ailment
from angr.ailment.block import Block
import networkx

from .utils import find_block_by_addr


class Goto:
    """
    Describe the existence of a goto (jump) statement. May have multiple gotos with the same address (targets
    will differ).
    """

    def __init__(self, src_addr, dst_addr, src_idx=None, dst_idx=None, src_ins_addr=None):
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.src_idx = src_idx
        self.dst_idx = dst_idx
        self.src_ins_addr = src_ins_addr

    def __hash__(self):
        return hash(f"{self.src_addr}{self.dst_addr}{self.src_idx}{self.dst_idx}")

    def __str__(self):
        if self.src_addr is None or self.dst_addr is None:
            return f"<Goto {self.__hash__()}>"

        src_idx_str = "" if self.src_idx is None else f".{self.src_idx}"
        dst_idx_str = "" if self.dst_idx is None else f".{self.dst_idx}"
        src_ins_addr_str = "" if self.src_ins_addr is None else f"{hex(self.src_ins_addr)}"
        return f"<Goto: [{hex(self.src_addr)}@{src_ins_addr_str}{src_idx_str}] -> {hex(self.dst_addr)}{dst_idx_str}>"

    def __repr__(self):
        return self.__str__()


class GotoManager:
    """
    Container class for all Gotos found in a function after decompilation structuring.
    This should be populated using GotoSimplifier.
    """

    def __init__(self, func, gotos=None):
        self.func = func
        self.gotos: set[Goto] = gotos or set()

        self._gotos_by_addr = None

    def __str__(self):
        return f"<GotoManager: func[{hex(self.func.addr)}] {len(self.gotos)} gotos>"

    def __repr__(self):
        return self.__str__()

    def gotos_in_block(self, block: ailment.Block) -> set[Goto]:
        gotos_found = set()
        for goto in self.gotos:
            if goto.src_addr == block.addr:
                gotos_found.add(goto)
            else:
                block_addrs = {stmt.tags["ins_addr"] for stmt in block.statements if "ins_addr" in stmt.tags}
                if goto.src_ins_addr in block_addrs:
                    gotos_found.add(goto)

        return gotos_found

    def is_goto_edge(self, src: ailment.Block, dst: ailment.Block):
        src_gotos = self.gotos_in_block(src)
        for goto in src_gotos:
            if goto.dst_addr == dst.addr:
                return True
            block_addrs = {stmt.tags["ins_addr"] for stmt in dst.statements if "ins_addr" in stmt.tags}
            if goto.dst_addr in block_addrs:
                return True

        return False

    def find_goto_edges(self, graph: networkx.DiGraph) -> list[tuple[Block, Block]]:
        """
        This function finds all edges that are _potential_ gotos in the graph.
        The gotos are not guaranteed to be correct, but they are an approximation based on how the Phoenix
        structuring algorithm will select edges from the graph to be gotos in structuring.
        """
        # first collect all simple destinations known by the goto managers
        dst_blocks = set()
        goto_edges = []
        for goto in self.gotos:
            try:
                dst_block = find_block_by_addr(graph, goto.dst_addr)
            except ValueError:
                continue

            try:
                src_block = find_block_by_addr(graph, goto.src_addr)
            except ValueError:
                src_block = None

            if src_block is None:
                # try the instruction addrs in the block to find the goto
                try:
                    src_block = find_block_by_addr(graph, goto.src_ins_addr, insn_addr=True)
                except ValueError:
                    src_block = None

            if src_block is not None and dst_block is not None:
                # if you found the source, we dont need to try later things on this dst
                goto_edges.append((src_block, dst_block))
            elif dst_block is not None:
                dst_blocks.add(dst_block)

        return goto_edges



from __future__ import annotations

from typing import Literal

from archinfo.arch_soot import SootAddressDescriptor

from .block_id import BlockID


CFGNODE_K = tuple[int, int]  # tuple[addr, size]
CFGENODE_K = tuple[BlockID, int, int]  # tuple[BlockID, size, looping times]
SOOTNODE_K = SootAddressDescriptor
K = CFGNODE_K | CFGENODE_K | SOOTNODE_K

CFG_ADDR_TYPES = Literal["int", "block_id", "soot"]

__all__ = ["CFGENODE_K", "CFGNODE_K", "CFG_ADDR_TYPES", "SOOTNODE_K", "K"]



from __future__ import annotations

from angr.protos import cfg_pb2
from angr.serializable import Serializable


class IndirectJumpType:
    """
    The type of an indirect jump or call.
    """

    Jumptable_AddressLoadedFromMemory = 0
    Jumptable_AddressComputed = 1
    Vtable = 3
    Unknown = 255


class JumptableInfo:
    """
    Describes a jump table or a vtable.
    """

    __slots__ = ("addr", "entries", "entry_size", "size")

    def __init__(self, addr: int | None, size: int, entry_size: int, entries: list[int]):
        self.addr = addr
        self.size = size
        self.entry_size = entry_size
        self.entries = entries

    def serialize_to_cmessage(self):
        cmsg = cfg_pb2.JumptableInfo()
        if self.addr is not None:
            cmsg.addr = self.addr
        cmsg.size = self.size
        cmsg.entry_size = self.entry_size
        cmsg.entries.extend(self.entries)
        return cmsg

    @classmethod
    def parse_from_cmessage(cls, cmsg) -> JumptableInfo:
        addr = cmsg.addr if cmsg.HasField("addr") else None
        return cls(addr, cmsg.size, cmsg.entry_size, list(cmsg.entries))


class IndirectJump(Serializable):
    """
    Describes an indirect jump or call site.
    """

    __slots__ = (
        "addr",
        "func_addr",
        "ins_addr",
        "jumpkind",
        "jumptable",
        "jumptables",
        "resolved_targets",
        "stmt_idx",
        "type",
    )

    def __init__(
        self,
        addr: int,
        ins_addr: int,
        func_addr: int,
        jumpkind: str,
        stmt_idx: int,
        resolved_targets: list[int] | None = None,
        jumptable: bool = False,
        jumptable_addr: int | None = None,
        jumptable_size: int | None = None,
        jumptable_entry_size: int | None = None,
        jumptable_entries: list[int] | None = None,
        type_: int | None = IndirectJumpType.Unknown,
    ):
        self.addr = addr
        self.ins_addr = ins_addr
        self.func_addr = func_addr
        self.jumpkind = jumpkind
        self.stmt_idx = stmt_idx
        self.resolved_targets = set() if resolved_targets is None else set(resolved_targets)
        self.jumptable = jumptable
        self.jumptables: list[JumptableInfo] = []
        if (
            jumptable_addr is not None
            and jumptable_size is not None
            and jumptable_entry_size is not None
            and jumptable_entries is not None
        ):
            self.add_jumptable(jumptable_addr, jumptable_size, jumptable_entry_size, jumptable_entries)
        self.type = type_

    def add_jumptable(
        self,
        addr: int | None,
        size: int,
        entry_size: int,
        entries: list[int],
        is_primary: bool = False,
    ) -> None:
        ji = JumptableInfo(addr, size, entry_size, entries)
        if is_primary:
            self.jumptables.insert(0, ji)
        else:
            self.jumptables.append(ji)

    # for compatibility convenience

    @property
    def jumptable_addr(self) -> int | None:
        if self.jumptables:
            return self.jumptables[0].addr
        return None

    @property
    def jumptable_size(self) -> int | None:
        if self.jumptables:
            return self.jumptables[0].size
        return None

    @property
    def jumptable_entry_size(self) -> int | None:
        if self.jumptables:
            return self.jumptables[0].entry_size
        return None

    @property
    def jumptable_entries(self) -> list[int] | None:
        if self.jumptables:
            return self.jumptables[0].entries
        return None

    @classmethod
    def _get_cmsg(cls):
        return cfg_pb2.IndirectJump()

    def serialize_to_cmessage(self):
        cmsg = self._get_cmsg()
        cmsg.addr = self.addr
        cmsg.ins_addr = self.ins_addr
        cmsg.func_addr = self.func_addr
        cmsg.jumpkind = self.jumpkind
        cmsg.stmt_idx = self.stmt_idx
        cmsg.resolved_targets.extend(self.resolved_targets)
        cmsg.jumptable = self.jumptable
        for jt in self.jumptables:
            cmsg.jumptables.append(jt.serialize_to_cmessage())
        cmsg.type = self.type if self.type is not None else IndirectJumpType.Unknown
        return cmsg

    @classmethod
    def parse_from_cmessage(cls, cmsg, **kwargs) -> IndirectJump:
        obj = cls(
            addr=cmsg.addr,
            ins_addr=cmsg.ins_addr,
            func_addr=cmsg.func_addr,
            jumpkind=cmsg.jumpkind,
            stmt_idx=cmsg.stmt_idx,
            resolved_targets=list(cmsg.resolved_targets),
            jumptable=cmsg.jumptable,
            type_=cmsg.type,
        )
        for jt_cmsg in cmsg.jumptables:
            ji = JumptableInfo.parse_from_cmessage(jt_cmsg)
            obj.jumptables.append(ji)
        return obj

    def __repr__(self):
        status = ""
        if self.jumptable or self.jumptable_entries:
            status = "vtable" if self.type == IndirectJumpType.Vtable else "jumptable"
            if self.jumptable_addr is not None:
                status += f"@{self.jumptable_addr:#08x}"
            if self.jumptable_entries is not None:
                status += f" with {len(self.jumptable_entries)} entries"
            if len(self.jumptables) > 1:
                status += f" (+{len(self.jumptables) - 1} jumptables)"

        return "<IndirectJump {:#08x} - ins {:#08x}{}>".format(self.addr, self.ins_addr, " " + status if status else "")



from __future__ import annotations
import angr


class recv(angr.SimProcedure):
    # pylint:disable=arguments-differ,unused-argument

    def run(self, fd, dst, length, flags):
        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1

        return simfd.read(dst, length)



from __future__ import annotations

from .plugin import SimStatePlugin
from .libc import SimStateLibc
from .inspect import SimInspector, NO_OVERRIDE, BP_BEFORE, BP_AFTER, BP_BOTH, BP_IPDB, BP_IPYTHON
from .posix import PosixDevFS, PosixProcFS, SimSystemPosix
from .solver import SimSolver
from .light_registers import SimLightRegisters
from .log import SimStateLog
from .history import SimStateHistory
from .scratch import SimStateScratch
from .cgc import SimStateCGC
from .gdb import GDB
from .uc_manager import SimUCManager
from .unicorn_engine import Unicorn
from .sim_action import SimAction, SimActionExit, SimActionConstraint, SimActionOperation, SimActionData
from .sim_action_object import SimActionObject
from .sim_event import SimEvent, resource_event
from .callstack import CallStack
from .globals import SimStateGlobals
from .preconstrainer import SimStatePreconstrainer
from .loop_data import SimStateLoopData
from .view import SimRegNameView, SimMemView, StructMode
from .filesystem import Stat, SimFilesystem, SimMount, SimHostFilesystem
from .heap import SimHeapBase, SimHeapBrk, SimHeapLibc, SimHeapPTMalloc, PTChunk, PTChunkIterator
from .jni_references import SimStateJNIReferences
from .javavm_classloader import SimJavaVmClassloader
from .symbolizer import SimSymbolizer
from .debug_variables import SimDebugVariable, SimDebugVariablePlugin
from .edge_hitmap import SimStateEdgeHitmap

__all__ = (
    "BP_AFTER",
    "BP_BEFORE",
    "BP_BOTH",
    "BP_IPDB",
    "BP_IPYTHON",
    "GDB",
    "NO_OVERRIDE",
    "CallStack",
    "PTChunk",
    "PTChunkIterator",
    "PosixDevFS",
    "PosixProcFS",
    "SimAction",
    "SimActionConstraint",
    "SimActionData",
    "SimActionExit",
    "SimActionObject",
    "SimActionOperation",
    "SimDebugVariable",
    "SimDebugVariablePlugin",
    "SimEvent",
    "SimFilesystem",
    "SimHeapBase",
    "SimHeapBrk",
    "SimHeapLibc",
    "SimHeapPTMalloc",
    "SimHostFilesystem",
    "SimInspector",
    "SimJavaVmClassloader",
    "SimLightRegisters",
    "SimMemView",
    "SimMount",
    "SimRegNameView",
    "SimSolver",
    "SimStateCGC",
    "SimStateEdgeHitmap",
    "SimStateGlobals",
    "SimStateHistory",
    "SimStateJNIReferences",
    "SimStateLibc",
    "SimStateLog",
    "SimStateLoopData",
    "SimStatePlugin",
    "SimStatePreconstrainer",
    "SimStateScratch",
    "SimSymbolizer",
    "SimSystemPosix",
    "SimUCManager",
    "Stat",
    "StructMode",
    "Unicorn",
    "resource_event",
)



from __future__ import annotations
from typing import Any, TYPE_CHECKING
import time
import json
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.exc import DatabaseError
from sqlalchemy.pool import NullPool

from angr.errors import AngrCorruptDBError, AngrIncompatibleDBError, AngrDBError
from angr.procedures import SIM_PROCEDURES
from angr.project import Project
from .models import Base, DbInformation
from .serializers import LoaderSerializer, KnowledgeBaseSerializer

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase


class AngrDB:
    """
    AngrDB provides a storage solution for an angr project, its knowledge bases, and some other types of data. It is
    designed to use an SQL-based database as the storage backend.
    """

    ALL_TABLES = [
        "objects",
    ]

    VERSION = 1

    def __init__(self, project=None, nullpool=False):
        self.project = project
        self.config = {}
        self.engine = None
        self._nullpool = nullpool

    @contextmanager
    def open_db(self, db_str="sqlite:///:memory:"):
        try:
            kwargs = {}
            if self._nullpool:
                kwargs["poolclass"] = NullPool
            self.engine = create_engine(db_str, **kwargs)
            Base.metadata.create_all(self.engine)
            Session = sessionmaker(bind=self.engine)
            yield Session
        except DatabaseError as ex:
            raise AngrCorruptDBError("The target file may not be an angr database or it is corrupted.") from ex
        except Exception as ex:
            raise AngrDBError from ex

    @staticmethod
    @contextmanager
    def session_scope(Session):
        session = Session()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    @staticmethod
    def save_info(session, key, value):
        """
        Save an information entry to the database.

        :param session:
        :param key:
        :param value:
        :return:
        """

        db_info = session.query(DbInformation).filter_by(key=key).scalar()
        if db_info is not None:
            db_info.value = value
        else:
            db_info = DbInformation(key=key, value=value)
            session.add(db_info)

    @staticmethod
    def get_info(session, key):
        """
        Get an information entry from the database.

        :param session:
        :param key:
        :return:
        """

        db_info = session.query(DbInformation).filter_by(key=key).scalar()
        if db_info is None:
            return None
        return db_info.value

    def update_dbinfo(self, session, extra_info: dict[str, str] | None = None):
        """
        Update the information in database.

        :param session:
        :return:
        """

        self.save_info(session, "version", str(self.VERSION))
        self.save_info(session, "saved_at", str(int(time.time())))

        if extra_info:
            for key, value in extra_info.items():
                self.save_info(session, str(key), str(value))

    def get_dbinfo(self, session, extra_info: dict[str, str] | None = None):
        """
        Get database information.

        :param session:
        :return:        A dict of information entries.
        """

        d = {}

        # version
        version = self.get_info(session, "version")
        if version is not None:
            version = int(version)
        d["version"] = version

        # saved_at
        saved_at = self.get_info(session, "saved_at")
        if saved_at is not None:
            saved_at = int(saved_at)
        d["saved_at"] = saved_at

        if extra_info is not None:
            # store *everything* into the dict
            for entry in session.query(DbInformation):
                extra_info[entry.key] = entry.value

        return d

    def db_compatible(self, version):
        """
        Checks if the given database version is compatible with the current AngrDB class.

        :param int version: The version of the database.
        :return:            True if compatible, False otherwise.
        :rtype:             bool
        """

        return version == self.VERSION

    def invalidate(self):
        if self.engine is not None:
            self.engine.dispose()

    def dump(self, db_path, kbs: list[KnowledgeBase] | None = None, extra_info: dict[str, Any] | None = None):
        assert self.project is not None

        db_str = f"sqlite:///{db_path}"

        with self.open_db(db_str) as Session, self.session_scope(Session) as session:
            # Dump the loader
            LoaderSerializer.dump(session, self.project.loader)
            # Dump the knowledge base

            if kbs is None:
                kbs = [self.project.kb]

            for kb in kbs:
                KnowledgeBaseSerializer.dump(session, kb)

            # Save unresolvable target hook addresses so CFG edges can be resolved after load. CFGFast
            # allocates pseudo-addresses for UnresolvableCallTarget/JumpTarget via get_pseudo_addr(),
            # which are baked into CFG node/edge addresses. These addresses are not reproducible on a
            # fresh loader, so we must save and restore them explicitly.
            call_target_addrs: list[int] = []
            jump_target_addrs: list[int] = []
            for addr, proc in self.project._sim_procedures.items():
                name = type(proc).__name__
                if name == "UnresolvableCallTarget":
                    call_target_addrs.append(addr)
                elif name == "UnresolvableJumpTarget":
                    jump_target_addrs.append(addr)
            if call_target_addrs:
                self.save_info(session, "unresolvable_call_target_addrs", json.dumps(call_target_addrs))
            if jump_target_addrs:
                self.save_info(session, "unresolvable_jump_target_addrs", json.dumps(jump_target_addrs))

            # Update the information
            self.update_dbinfo(session, extra_info=extra_info)

    def load(
        self,
        db_path: str,
        kb_names: list[str] | None = None,
        other_kbs: dict[str, KnowledgeBase] | None = None,
        extra_info: dict[str, Any] | None = None,
    ):
        db_str = f"sqlite:///{db_path}"

        with self.open_db(db_str) as Session, self.session_scope(Session) as session:
            # Compatibility check
            dbinfo = self.get_dbinfo(session, extra_info=extra_info)
            if not self.db_compatible(dbinfo.get("version", None)):
                raise AngrIncompatibleDBError(
                    "Version {} is incompatible with the current version of angr.".format(dbinfo.get("version", None))
                )

            # Load the loader
            loader = LoaderSerializer.load(session)
            # Create the project
            proj = Project(loader)

            # Restore unresolvable target hooks at the exact addresses used by the original CFGFast
            # analysis. These addresses are embedded in CFG edges and must be hooked with the correct
            # SimProcedure for Clinic's indirect call resolution to work.
            call_addrs_str = self.get_info(session, "unresolvable_call_target_addrs")
            if call_addrs_str:
                try:
                    call_addrs_list = json.loads(call_addrs_str)
                    if not isinstance(call_addrs_list, list):
                        raise TypeError("Expected a list of addresses")
                    for addr in call_addrs_list:
                        if isinstance(addr, int):
                            proj.hook(addr, SIM_PROCEDURES["stubs"]["UnresolvableCallTarget"](), replace=True)
                except (json.JSONDecodeError, TypeError):
                    pass

            jump_addrs_str = self.get_info(session, "unresolvable_jump_target_addrs")
            if jump_addrs_str:
                try:
                    jump_addrs_list = json.loads(jump_addrs_str)
                    if not isinstance(jump_addrs_list, list):
                        raise json.JSONDecodeError("Expected a list of addresses", "dummy", 0)
                    for addr in jump_addrs_list:
                        if isinstance(addr, int):
                            proj.hook(addr, SIM_PROCEDURES["stubs"]["UnresolvableJumpTarget"](), replace=True)
                except (json.JSONDecodeError, TypeError):
                    pass

            if kb_names is None:
                kb_names = ["global"]

            if (len(kb_names) != 1 or kb_names[0] != "global") and other_kbs is None:
                raise ValueError(
                    'You must provide a dict via "other_kbs" to collect angr KnowledgeBases '
                    "that are not the global one."
                )

            # Load knowledgebases
            for kb_name in kb_names:
                kb = KnowledgeBaseSerializer.load(session, proj, kb_name)
                if kb is not None:
                    if kb_name == "global":
                        proj.kb = kb
                    else:
                        other_kbs[kb_name] = kb

            return proj



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from typing import Any

from angr.ailment import Block
from angr.ailment.statement import Assignment, Label, ConditionalJump
from angr.ailment.expression import BinaryOp, Call, VirtualVariable, Const, Phi, Load, Expression
from angr.analyses.decompiler.optimization_passes.optimization_pass import OptimizationPass, OptimizationPassStage
from angr.utils.ail import is_phi_assignment
from angr.utils.bits import u2s


class InlinedStrlenSimplifier(OptimizationPass):
    """
    Abstracts inlined strlen functions into strlen() calls.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Identify and simplify inlined strlen() functions"
    DESCRIPTION = "Identify and simplify inlined strlen() functions"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.analyze()

    def _check(self):
        strlen_descs = self._find_strlen_patterns()

        return bool(strlen_descs), {"descs": strlen_descs}

    def _analyze(self, cache=None):
        if not cache or "descs" not in cache:
            return

        blocks_dict = {(block.addr, block.idx): block for block in self._graph}

        for desc in cache["descs"]:
            self._convert_block_to_strlen(
                blocks_dict[(desc["block_addr"], desc["block_idx"])], desc["str"], desc["result_var"], desc["load_size"]
            )

    def _convert_block_to_strlen(
        self, block: Block, str_expr: Expression, result_var: VirtualVariable, load_size: int
    ) -> None:
        """
        Convert the given block into a strlen() call.
        """

        match load_size:
            case 1:
                func_name = "strlen"
            case 2:
                func_name = "wcslen"
            case _:
                raise RuntimeError(f"Unsupported load size {load_size} for strlen simplification.")
        strlen_call = Call(None, func_name, args=[str_expr], bits=result_var.bits, ins_addr=block.addr)
        strlen_stmt = Assignment(None, result_var, strlen_call, ins_addr=block.addr)
        new_block = Block(block.addr, block.original_size, statements=[strlen_stmt], idx=block.idx)
        self._update_block(block, new_block)

        # remove the self-loop edge
        assert self.out_graph is not None
        if self.out_graph.has_edge(new_block, new_block):
            self.out_graph.remove_edge(new_block, new_block)

    def _find_strlen_patterns(self):
        """
        Identify strlen patterns.
        """

        blocks_dict = {(block.addr, block.idx): block for block in self._graph}
        strlen_descs = []

        for block in self._graph:
            r, d = self._is_strlen_pattern_1_core_block(block)
            if r:
                assert d is not None
                if d["load_size"] in {1, 2}:
                    # ensure the index variable is initialized to -1
                    index_var_src: tuple[int, int | None] = d["index_var_src"]
                    index_var: VirtualVariable = d["index_var"]

                    index_var_init_value = self._find_var_init_value(blocks_dict[index_var_src], index_var)
                    if (
                        index_var_init_value is not None
                        and isinstance(index_var_init_value, Const)
                        and u2s(index_var_init_value.value_int, index_var.bits) == -1
                    ):
                        strlen_descs.append(d)
        return strlen_descs

    @staticmethod
    def _is_strlen_pattern_1_core_block(block: Block) -> tuple[bool, dict[str, Any] | None]:
        """
        Pattern 1:

        ## Block 1400145ee
        00 | 0x1400145ee | vvar_403{r16|8b} = 𝜙@64b [((5368792551, None), vvar_1507{r16|8b}), ((5368792558, None),
                vvar_1509{r16|8b})]
        01 | 0x1400145ee | LABEL_1400145ee:
        02 | 0x1400145ee | vvar_1509{r16|8b} = (vvar_403{r16|8b} + 0x1<64>)
        03 | 0x1400145f4 | if ((Load(addr=(vvar_394{r40|8b} + vvar_1509{r16|8b}), size=1, endness=Iend_LE) == 0x0<8>))
                { Goto 0x1400145f6<64> } else { Goto 0x1400145ee<64> }

        this method only verifies if the given block follows the pattern of the strlen core block; it does not verify
        if the index variable is properly initialized to -1.
        """

        phi_stmt: Assignment | None = None
        assignment_stmt: Assignment | None = None
        cond_jump: ConditionalJump | None = None
        for stmt in block.statements:
            if is_phi_assignment(stmt):
                if phi_stmt is not None:
                    return False, None
                phi_stmt = stmt  # type: ignore
            elif isinstance(stmt, Label):
                continue
            elif isinstance(stmt, Assignment):
                if assignment_stmt is not None:
                    return False, None
                assignment_stmt = stmt
            elif isinstance(stmt, ConditionalJump):
                if cond_jump is not None:
                    return False, None
                cond_jump = stmt
            else:
                return False, None

        if phi_stmt is None or assignment_stmt is None or cond_jump is None:
            return False, None

        v0 = phi_stmt.dst

        # check the assignment statement
        v1 = assignment_stmt.dst
        if isinstance(assignment_stmt.src, BinaryOp) and assignment_stmt.src.op == "Add":
            binop: BinaryOp = assignment_stmt.src
            if (
                isinstance(binop.operands[0], VirtualVariable)
                and binop.operands[0].likes(v0)
                and isinstance(binop.operands[1], Const)
                and binop.operands[1].value == 1
            ) or (
                isinstance(binop.operands[1], VirtualVariable)
                and binop.operands[1].likes(v0)
                and isinstance(binop.operands[0], Const)
                and binop.operands[0].value == 1
            ):
                pass
            else:
                return False, None
        else:
            return False, None

        if v0.likes(v1):
            # unexpected due to SSA but still gotta check
            return False, None

        # check the phi statement
        v_phi_src_var: VirtualVariable | None = None
        v_phi_src: tuple[int, int | None] | None = None
        assert isinstance(phi_stmt.src, Phi)
        if len(phi_stmt.src.src_and_vvars) == 2:
            src_to_vvars = dict(phi_stmt.src.src_and_vvars)
            if (block.addr, block.idx) not in src_to_vvars:
                return False, None
            vvar = src_to_vvars[(block.addr, block.idx)]
            if vvar is not None and vvar.likes(v1):
                # self-loop
                v_phi_src, v_phi_src_var = next(
                    (src, vvar) for src, vvar in phi_stmt.src.src_and_vvars if src != (block.addr, block.idx)
                )
            else:
                return False, None
        else:
            return False, None

        assert v_phi_src is not None and v_phi_src_var is not None

        # check the conditional jump
        cond = cond_jump.condition
        if (
            isinstance(cond, BinaryOp)
            and cond.op in {"CmpEQ", "CmpNE"}
            and isinstance(cond.operands[0], Load)
            and isinstance(cond.operands[1], Const)
            and cond.operands[1].value == 0
        ):
            load_addr = cond.operands[0].addr
            load_size = cond.operands[0].size
            str_src = None
            r = False
            if (
                isinstance(load_addr, BinaryOp)
                and load_addr.op == "Add"
                and any(op.likes(v1) for op in load_addr.operands)
                and load_size == 1
            ):
                str_src = next(op for op in load_addr.operands if not op.likes(v1))
                if (
                    cond.op == "CmpEQ"
                    and isinstance(cond_jump.false_target, Const)
                    and cond_jump.false_target.value == block.addr
                    and cond_jump.false_target_idx == block.idx
                    and isinstance(cond_jump.true_target, Const)
                    and (cond_jump.true_target.value, cond_jump.true_target_idx) != (block.addr, block.idx)
                ) or (
                    cond.op == "CmpNE"
                    and isinstance(cond_jump.true_target, Const)
                    and cond_jump.true_target.value == block.addr
                    and cond_jump.true_target_idx == block.idx
                    and isinstance(cond_jump.false_target, Const)
                    and (cond_jump.false_target.value, cond_jump.false_target_idx) != (block.addr, block.idx)
                ):
                    r = True

            elif (
                isinstance(load_addr, BinaryOp)
                and load_addr.op == "Add"
                and isinstance(load_addr.operands[1], BinaryOp)
                and load_addr.operands[1].op == "Mul"
            ):
                mul_op: BinaryOp = load_addr.operands[1]
                if (
                    isinstance(mul_op.operands[0], VirtualVariable)
                    and mul_op.operands[0].likes(v1)
                    and isinstance(mul_op.operands[1], Const)
                    and mul_op.operands[1].value == load_size
                ):
                    str_src = load_addr.operands[0]
                    r = False
                    if (
                        cond.op == "CmpEQ"
                        and isinstance(cond_jump.false_target, Const)
                        and cond_jump.false_target.value == block.addr
                        and cond_jump.false_target_idx == block.idx
                        and isinstance(cond_jump.true_target, Const)
                        and (cond_jump.true_target.value, cond_jump.true_target_idx) != (block.addr, block.idx)
                    ) or (
                        cond.op == "CmpNE"
                        and isinstance(cond_jump.true_target, Const)
                        and cond_jump.true_target.value == block.addr
                        and cond_jump.true_target_idx == block.idx
                        and isinstance(cond_jump.false_target, Const)
                        and (cond_jump.false_target.value, cond_jump.false_target_idx) != (block.addr, block.idx)
                    ):
                        r = True

            if r:
                return True, {
                    "block_addr": block.addr,
                    "block_idx": block.idx,
                    "str": str_src,
                    "index_var": v_phi_src_var,
                    "index_var_src": v_phi_src,
                    "load_size": load_size,
                    "result_var": v1,
                }

        return False, None

    @staticmethod
    def _find_var_init_value(block: Block, var: VirtualVariable) -> Expression | None:
        """
        Find the initialization value of a variable in a block.
        """

        for stmt in block.statements:
            if isinstance(stmt, Assignment) and stmt.dst.likes(var):
                return stmt.src
        return None



# pylint:disable=arguments-differ
from __future__ import annotations

from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class ISPOMixin(MemoryMixin):
    """
    An implementation of the International Stateless Persons Organisation, a mixin which should be applied as a bottom
    layer for memories which have no state and must redirect certain operations to a parent memory. Main usecase is
    for memory region classes which are stored within other memories, such as pages.
    """

    def set_state(self, state):
        raise Exception("Cannot set state on this stateless object")

    def _default_value(self, *args, memory, **kwargs):
        try:
            func = memory._default_value
        except AttributeError as e:
            raise Exception("memory kwarg must be passed to this stateless object") from e
        else:
            return func(*args, **kwargs)

    def _add_constraints(self, *args, memory, **kwargs):
        try:
            func = memory._add_constraints
        except AttributeError as e:
            raise Exception("memory kwarg must be passed to this stateless object") from e
        else:
            return func(*args, **kwargs)

    def _merge_values(self, *args, memory, **kwargs):
        try:
            func = memory._merge_values
        except AttributeError as ex:
            raise Exception("memory kwarg must be passed to this stateless object") from ex
        else:
            return func(*args, **kwargs)

    def _merge_labels(self, *args, memory, **kwargs):
        try:
            func = memory._merge_labels
        except AttributeError as ex:
            raise Exception("memory kwarg must be passed to this stateless object") from ex
        else:
            return func(*args, **kwargs)

    def _update_mappings(self, *args, memory, **kwargs):
        try:
            func = memory._update_mappings
        except AttributeError as ex:
            raise Exception("memory kwarg must be passed to this stateless object") from ex
        else:
            return func(*args, **kwargs)



from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from collections import Counter
from angr.analyses.decompiler.structured_codegen.c import CStructuredCodeWalker
from angr.analyses.analysis import Analysis

if TYPE_CHECKING:
    from angr.analyses.decompiler.structured_codegen.c import CFunction

_l = logging.getLogger(name=__name__)


class ScopeOpsWalker(CStructuredCodeWalker):
    """
    ScopeOpsWalker walks a C construct (typically a C function) and extracts counts of all operations in each scope.
    Its intended use is to reason about the intent of different parts of functions (e.g., crypto implementations).
    """

    def __init__(self):
        self.current_addr = None
        self.found_ops = {}

    def handle(self, obj):
        if type(obj).__name__ in [
            "CFunction",
            "CWhileLoop",
            "CDoWhileLoop",
            "CSwitchCase",
            "CIfBreak",
            "CForLoop",
            "CIfElse",
        ]:
            old_addr = self.current_addr
            self.current_addr = obj.tags.get("ins_addr", getattr(obj, "addr", None))
            self.found_ops.setdefault(self.current_addr, Counter())
            super().handle(obj)
            self.current_addr = old_addr
            return self.found_ops
        return super().handle(obj)

    def handle_CBinaryOp(self, obj):
        self.found_ops[self.current_addr][obj.op] += 1
        return self.found_ops

    def handle_CUnaryOp(self, obj):
        self.found_ops[self.current_addr][obj.op] += 1
        return self.found_ops


class ScopeOpsAnalyzer(Analysis):
    """
    An analysis that extracts and analyzes operations used by different scopes of a function.
    """

    def __init__(self, cfunc: CFunction):
        self._cfunc = cfunc
        self.scope_ops = ScopeOpsWalker().handle(cfunc)

    def filter_scopes(self, wanted_ops: set, min_count):
        return [
            a for a, s in self.scope_ops.items() if sum(self.scope_ops[a][v] for v in s if v in wanted_ops) >= min_count
        ]

    def crypto_scopes(self, min_count=20):
        return self.filter_scopes({"Xor", "Ror", "Rol", "Shl", "Shr"}, min_count)



# This submodule stores and manages FLIRT signatures
from __future__ import annotations

import os
import json
from collections import defaultdict
import logging

from angr.analyses.flirt import (
    FlirtSignature,
    FlirtSignatureParsed,
    FlirtSignatureError,
    flirt_arch_to_arch_name,
    flirt_os_type_to_os_name,
)

_l = logging.getLogger(__name__)


FS = FlirtSignature

# A dict from architecture names to FLIRT signatures under that architecture. Arch names are always in lower case.
FLIRT_SIGNATURES_BY_ARCH: dict[str, list[FlirtSignature]] = defaultdict(list)
LIBRARY_TO_SIGNATURES: dict[str, list[FlirtSignature]] = defaultdict(list)
STRING_TO_LIBRARIES: dict[str, set[str]] = defaultdict(set)


def load_signature(sig_path: str, meta_path: str | None = None) -> tuple[str, FlirtSignature] | None:
    """
    Load a single FLIRT signature from a specific path.

    :param sig_path:    Location of the FLIRT signature.
    :return:            A FlirtSignature object if loading was successful, None otherwise.
    """

    # parse it
    try:
        with open(sig_path, "rb") as f:
            sig_parsed = FlirtSignatureParsed.parse(f)
    except FlirtSignatureError:
        return None

    # is there a meta data file?
    if meta_path is not None and os.path.isfile(meta_path):
        # yes!
        with open(meta_path) as f:
            meta = json.load(f)

        arch = str(meta.get("arch", "Unknown"))
        platform = str(meta.get("platform", "UnknownOS"))
        os_name = meta.get("os", None)
        os_version = meta.get("os_version", None)
        compiler = meta.get("compiler", None)
        compiler_version = meta.get("compiler_version", None)
        unique_strings = meta.get("unique_strings", None)

    else:
        # nope... we need to extract information from the signature file
        arch = flirt_arch_to_arch_name(sig_parsed.arch, sig_parsed.app_types)
        platform = flirt_os_type_to_os_name(sig_parsed.os_types)
        os_name = None
        os_version = None
        unique_strings = None
        compiler = None
        compiler_version = None

    signature = FlirtSignature(
        arch,
        platform,
        sig_parsed.libname,
        sig_path,
        unique_strings=unique_strings,
        compiler=compiler,
        compiler_version=compiler_version,
        os_name=os_name,
        os_version=os_version,
    )

    return arch, signature


def load_signatures(path: str) -> None:
    """
    Recursively load all FLIRT signatures under a specific path.

    :param path:    Location of FLIRT signatures.
    """

    FLIRT_SIGNATURES_BY_ARCH.clear()
    LIBRARY_TO_SIGNATURES.clear()
    STRING_TO_LIBRARIES.clear()

    for root, _, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(".sig"):
                sig_path = os.path.join(root, filename)
                meta_path = os.path.join(root, filename[:-4] + ".meta")
                r = load_signature(sig_path, meta_path=meta_path)
                if r is None:
                    _l.warning("Failed to load FLIRT signature file %s.", sig_path)
                    continue

                arch, signature = r
                FLIRT_SIGNATURES_BY_ARCH[arch].append(signature)

    # fill in LIBRARY_TO_SIGNATURES and STRING_TO_LIBRARIES
    for sigs in FLIRT_SIGNATURES_BY_ARCH.values():
        for sig in sigs:
            LIBRARY_TO_SIGNATURES[sig.sig_name].append(sig)
            if sig.unique_strings:
                for us in sig.unique_strings:
                    STRING_TO_LIBRARIES[us].add(sig.sig_name)


__all__ = (
    "FLIRT_SIGNATURES_BY_ARCH",
    "FS",
    "LIBRARY_TO_SIGNATURES",
    "STRING_TO_LIBRARIES",
    "FlirtSignature",
    "load_signature",
    "load_signatures",
)



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Semantic variable naming for pointer variables.

This module detects variables used as pointers and names them appropriately
(e.g., ptr, p, cur, next, prev).
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import logging
from collections import defaultdict

from angr import ailment
from angr.ailment.expression import BinaryOp, Call, UnaryOp, Const, Load
from angr.ailment.statement import Assignment, Store, SideEffectStatement
from angr.sim_variable import SimVariable

from .naming_base import ClinicNamingBase

if TYPE_CHECKING:
    from angr.ailment import Expression

l = logging.getLogger(name=__name__)

# Names for pointer variables based on usage pattern
POINTER_NAMES = {
    "generic": ["ptr", "p", "addr"],
    "iterator": ["cur", "iter", "node"],
    "linked_list": ["next", "prev", "head", "tail"],
    "string": ["str", "s", "buf"],
    "array": ["arr", "data", "base"],
}

# Functions that return pointers
POINTER_RETURNING_FUNCTIONS = {
    "malloc",
    "calloc",
    "realloc",
    "alloca",
    "strdup",
    "strndup",
    "strchr",
    "strrchr",
    "strstr",
    "strpbrk",
    "memchr",
    "memmem",
    "fopen",
    "fdopen",
    "tmpfile",
    "opendir",
    "dlopen",
    "dlsym",
    "mmap",
    "fgets",
    "gets",
}

# Functions that take pointer parameters (param_index -> is_output)
POINTER_PARAM_FUNCTIONS = {
    "memcpy": {0: True, 1: False},  # dst, src
    "memmove": {0: True, 1: False},
    "memset": {0: True},
    "strcpy": {0: True, 1: False},
    "strncpy": {0: True, 1: False},
    "strcat": {0: True, 1: False},
    "strcmp": {0: False, 1: False},
    "strncmp": {0: False, 1: False},
    "strlen": {0: False},
    "free": {0: False},
    "fread": {0: True},
    "fwrite": {0: False},
    "fprintf": {0: False},
    "sprintf": {0: True},
    "snprintf": {0: True},
}


class PointerNaming(ClinicNamingBase):
    """
    Detects variables used as pointers and renames them.

    Pointer patterns detected:
    - Variables used as addresses in Load/Store operations
    - Variables involved in pointer arithmetic
    - Variables passed to functions expecting pointers
    - Variables receiving pointer-returning function results
    - Variables used in linked-list patterns (next/prev access)
    """

    PRIORITY = 45  # After loop counters, before array indices

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pointer_candidates: dict[SimVariable, dict] = {}
        self._name_counter = defaultdict(int)  # category -> count

    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze the graph for pointer variable patterns.

        :return: Dictionary mapping SimVariable to new name
        """
        if not self._graph:
            return {}

        # Find variables used as pointers in various patterns
        self._find_dereference_pointers()
        self._find_pointer_arithmetic()
        self._find_function_pointer_params()
        self._find_linked_list_patterns()

        # Score and assign names
        self._assign_pointer_names()

        return self._var_to_new_name

    def _find_dereference_pointers(self) -> None:
        """Find variables used as addresses in Load/Store operations."""
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                # Check Store statements
                if isinstance(stmt, Store):
                    self._check_address_expr(stmt.addr, "store")

                # Check assignments with Load on the right side
                if isinstance(stmt, Assignment):
                    self._check_expr_for_loads(stmt.src)
                    # Also check if dst is a dereference
                    if isinstance(stmt.dst, Load):
                        self._check_address_expr(stmt.dst.addr, "store")

    def _check_expr_for_loads(self, expr) -> None:
        """Recursively check expression for Load operations."""
        if expr is None:
            return

        if isinstance(expr, Load):
            self._check_address_expr(expr.addr, "load")
            return

        if isinstance(expr, BinaryOp):
            for operand in expr.operands:
                self._check_expr_for_loads(operand)
        elif isinstance(expr, UnaryOp):
            self._check_expr_for_loads(expr.operand)

    def _check_address_expr(self, addr_expr: Expression, access_type: str) -> None:
        """Check an address expression for pointer variables."""
        # Direct variable as address
        var = self._get_linked_variable(addr_expr)
        if var is not None:
            self._record_pointer(var, "dereference", access_type)
            return

        # Pointer arithmetic: ptr + offset or ptr - offset
        if isinstance(addr_expr, BinaryOp) and addr_expr.op in ("Add", "Sub"):
            op0, op1 = addr_expr.operands

            # Check if one operand is a variable (the pointer)
            var0 = self._get_linked_variable(op0)
            var1 = self._get_linked_variable(op1)

            if var0 is not None and (isinstance(op1, Const) or self._is_index_like(op1)):
                self._record_pointer(var0, "arithmetic", access_type)
            elif var1 is not None and isinstance(op0, Const):
                self._record_pointer(var1, "arithmetic", access_type)

    @staticmethod
    def _is_index_like(expr: Expression) -> bool:
        """
        Check if expression looks like an index (e.g., i * 4).
        """
        if isinstance(expr, BinaryOp) and expr.op == "Mul":
            op0, op1 = expr.operands
            return isinstance(op0, Const) or isinstance(op1, Const)
        return False

    def _find_pointer_arithmetic(self) -> None:
        """
        Find variables involved in pointer arithmetic patterns.
        """
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if not isinstance(stmt, Assignment):
                    continue

                dst_var = self._get_linked_variable(stmt.dst)
                if dst_var is None:
                    continue

                src = stmt.src

                # Pattern: ptr = ptr + offset (pointer increment)
                if isinstance(src, BinaryOp) and src.op in ("Add", "Sub"):
                    op0, _ = src.operands
                    src_var = self._get_linked_variable(op0)

                    if src_var == dst_var:
                        # Self-increment: ptr = ptr + something
                        self._record_pointer(dst_var, "increment", "write")

                # Pattern: ptr = other_ptr (pointer copy/assignment)
                src_var = self._get_linked_variable(src)
                if src_var is not None and src_var in self._pointer_candidates:
                    # Copying from a known pointer
                    self._record_pointer(dst_var, "copy", "write")

    def _find_function_pointer_params(self) -> None:
        """Find variables passed as pointer parameters to functions."""
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if isinstance(stmt, SideEffectStatement):
                    self._analyze_call_for_pointers(stmt.expr, ret_expr=stmt.ret_expr)
                elif isinstance(stmt, Assignment) and isinstance(stmt.src, Call):
                    self._analyze_call_for_pointers(stmt.src, ret_expr=stmt.dst)

    def _analyze_call_for_pointers(self, call: Call, ret_expr=None) -> None:
        """Analyze a function call for pointer parameters and return values."""
        func_name = self._get_function_name(call)

        # Check return value
        if ret_expr is not None:
            ret_var = self._get_linked_variable(ret_expr)
            if ret_var is not None and func_name and self._normalize_name(func_name) in POINTER_RETURNING_FUNCTIONS:
                self._record_pointer(ret_var, "return_value", "write")

        # Check parameters
        if call.args is None:
            return

        if func_name:
            normalized = self._normalize_name(func_name)
            if normalized in POINTER_PARAM_FUNCTIONS:
                param_info = POINTER_PARAM_FUNCTIONS[normalized]
                for idx, is_output in param_info.items():
                    if idx < len(call.args):
                        var = self._get_linked_variable(call.args[idx])
                        if var is not None:
                            access = "write" if is_output else "read"
                            self._record_pointer(var, "func_param", access)

    def _find_linked_list_patterns(self) -> None:
        """Find variables used in linked-list traversal patterns."""
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                if not isinstance(stmt, Assignment):
                    continue

                dst_var = self._get_linked_variable(stmt.dst)
                if dst_var is None:
                    continue

                # Pattern: cur = cur->next (or similar field access)
                # This shows up as: cur = *(cur + offset)
                if isinstance(stmt.src, Load):
                    addr = stmt.src.addr
                    if isinstance(addr, BinaryOp) and addr.op == "Add":
                        op0, op1 = addr.operands
                        src_var = self._get_linked_variable(op0)

                        if src_var == dst_var and isinstance(op1, Const):
                            # Self-referential: cur = *(cur + offset)
                            # This is a classic linked-list traversal pattern
                            self._record_pointer(dst_var, "linked_list", "iterator")

    def _record_pointer(self, var: SimVariable, pattern: str, access: str) -> None:
        """Record a variable as being used as a pointer."""
        if var not in self._pointer_candidates:
            self._pointer_candidates[var] = {
                "score": 0,
                "patterns": set(),
                "accesses": set(),
                "is_iterator": False,
                "is_linked_list": False,
            }

        info = self._pointer_candidates[var]
        info["patterns"].add(pattern)
        info["accesses"].add(access)

        # Score based on pattern
        pattern_scores = {
            "dereference": 25,
            "arithmetic": 20,
            "increment": 15,
            "func_param": 15,
            "return_value": 20,
            "copy": 10,
            "linked_list": 30,
        }
        info["score"] += pattern_scores.get(pattern, 10)

        # Mark special types
        if pattern == "linked_list":
            info["is_linked_list"] = True
        if pattern == "increment" or access == "iterator":
            info["is_iterator"] = True

    def _assign_pointer_names(self) -> None:
        """Assign names to pointer variables."""
        # Sort by score (highest first)
        sorted_candidates = sorted(self._pointer_candidates.items(), key=lambda x: (-x[1]["score"], str(x[0].ident)))

        # Threshold for considering a variable as a pointer
        SCORE_THRESHOLD = 20

        for var, info in sorted_candidates:
            if info["score"] < SCORE_THRESHOLD:
                continue

            # Choose name based on usage pattern
            category = "iterator" if info["is_linked_list"] or info["is_iterator"] else "generic"

            names = POINTER_NAMES[category]
            idx = self._name_counter[category]

            new_name = names[idx] if idx < len(names) else f"ptr{sum(self._name_counter.values())}"
            self._var_to_new_name[var] = new_name
            self._name_counter[category] += 1

            l.debug("Identified pointer %s (score=%d, patterns=%s)", var.name, info["score"], info["patterns"])



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Sequence

if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Callable
    import networkx
    from angr import ailment


class RemoveNodeNotice(Exception):
    pass


class AILGraphWalker:
    """
    Walks an AIL graph and optionally replaces each node with a new node.
    """

    def __init__(
        self,
        graph: networkx.DiGraph[ailment.Block],
        handler: Callable[[ailment.Block], ailment.Block | None],
        replace_nodes: bool = False,
        strict_order_start: Sequence[ailment.Block] = (),
    ):
        self.graph = graph
        self.handler = handler
        self._replace_nodes = replace_nodes
        self._strict_order_start = strict_order_start
        self._edits: dict[ailment.Block, ailment.Block | None] = {}

    def _handle(self, node: ailment.Block):
        try:
            r = self.handler(node)
        except RemoveNodeNotice:
            # we need to remove this node
            self._edits[node] = None
        else:
            if r is not None and r is not node and self._replace_nodes:
                self._edits[node] = r

    def walk(self):
        if not self._strict_order_start:
            for node in self.graph.nodes():
                self._handle(node)
        else:
            traverse_in_order(self.graph, self._strict_order_start, self._handle)

        if self._replace_nodes:
            for old, new in self._edits.items():
                if new is None:
                    self.graph.remove_node(old)
                    continue

                in_edges = list(self.graph.in_edges(old, data=True))
                out_edges = list(self.graph.out_edges(old, data=True))

                self.graph.remove_node(old)
                self.graph.add_node(new)

                for src, _, data in in_edges:
                    if src is old:
                        self.graph.add_edge(new, new, **data)
                    else:
                        self.graph.add_edge(src, new, **data)

                for _, dst, data in out_edges:
                    if dst is old:
                        self.graph.add_edge(new, new, **data)
                    else:
                        self.graph.add_edge(new, dst, **data)


def traverse_in_order(
    ail_graph: networkx.DiGraph[ailment.Block],
    entry_blocks: Sequence[ailment.Block],
    visitor: Callable[[ailment.Block], Any],
):
    seen = set()
    pending = list(entry_blocks)
    last_pending = set(pending)
    forcing = set()

    # walk this graph in a special order to make sure we see defs of variables before their uses when possible
    while pending:
        stack = pending
        pending = set()

        while stack:
            block = stack.pop()
            if block in seen:
                continue
            if block in forcing or all(pred in seen for pred in ail_graph.pred[block]):
                # process it!
                visitor(block)
                seen.add(block)

                news = set(ail_graph.succ[block])
                news -= seen
                stack.extend(sorted(news))
            else:
                pending.add(block)

        if last_pending == pending:
            if len(pending) > 1:
                # emergency: break any ties by peeking ahead and seeing if one might unlock the other
                frontiers = {node: ({node}, {node}) for node in pending}
                while True:
                    for node, (seen2, frontier) in frontiers.items():
                        nxt = {n for f in frontier for n in ail_graph.succ[f] if n not in seen2}
                        seen2.update(nxt)
                        frontiers[node] = (seen2, nxt)
                    ready = {n for n, (_, frontier) in frontiers.items() if pending - frontier}
                    if ready:
                        break
                    if all(not frontier for _, frontier in frontiers.values()):
                        ready = pending
                        break
            else:
                ready = pending
            forcing = set(ready)

        last_pending = set(pending)
        pending = sorted(pending)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
import logging
from typing import TYPE_CHECKING
from collections.abc import Iterable, Mapping

from angr.ailment.manager import Manager
from angr.ailment.statement import Statement, Assignment, SideEffectStatement, Store, Jump
from angr.ailment.expression import Call, Tmp, Load, Const, Register, Convert, Expression, VirtualVariable
from angr.ailment import AILBlockViewer
from angr.code_location import AILCodeLocation
from angr.knowledge_plugins.key_definitions import atoms
from angr.analyses.s_propagator import SPropagatorAnalysis
from angr.analyses.s_reaching_definitions import SReachingDefinitionsAnalysis, SRDAModel
from angr.analyses import Analysis, register_analysis
from angr.utils.ssa import has_reference_to_vvar
from .peephole_optimizations import (
    MULTI_STMT_OPTS,
    STMT_OPTS,
    EXPR_OPTS,
    PeepholeOptimizationStmtBase,
    PeepholeOptimizationExprBase,
    PeepholeOptimizationMultiStmtBase,
)
from .utils import peephole_optimize_exprs, peephole_optimize_stmts, peephole_optimize_multistmts

if TYPE_CHECKING:
    from angr.ailment.block import Block


_l = logging.getLogger(name=__name__)


class HasCallExprWalker(AILBlockViewer):
    """
    Test if an expression contains a call expression inside.
    """

    def __init__(self):
        super().__init__()
        self.has_call_expr = False

    def _handle_SideEffectStatement(self, stmt_idx: int, stmt: SideEffectStatement, block: Block | None):  # pylint:disable=unused-argument
        self.has_call_expr = True

    def _handle_CallExpr(  # pylint:disable=unused-argument
        self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        self.has_call_expr = True


class BlockSimplifier(Analysis):
    """
    Simplify an AIL block.
    """

    def __init__(
        self,
        block: Block | None,
        ail_manager: Manager,
        func_addr: int | None = None,
        stack_pointer_tracker=None,
        peephole_optimizations: None
        | (
            Iterable[
                type[PeepholeOptimizationStmtBase]
                | type[PeepholeOptimizationExprBase]
                | type[PeepholeOptimizationMultiStmtBase]
            ]
        ) = None,
        preserve_vvar_ids: set[int] | None = None,
        type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]] | None = None,
        cached_reaching_definitions=None,
        cached_propagator=None,
    ):
        """
        :param block:   The AIL block to simplify. Setting it to None to skip calling self._analyze(), which is useful
                        in test cases.
        """

        self.block = block
        self.func_addr = func_addr

        self._stack_pointer_tracker = stack_pointer_tracker
        self._preserve_vvar_ids = preserve_vvar_ids
        self._type_hints = type_hints
        self._ail_manager = ail_manager

        if peephole_optimizations is None:
            self._expr_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in EXPR_OPTS
            ]
            self._stmt_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in STMT_OPTS
            ]
            self._multistmt_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in MULTI_STMT_OPTS
            ]
        else:
            self._expr_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in peephole_optimizations
                if issubclass(cls, PeepholeOptimizationExprBase)
            ]
            self._stmt_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in peephole_optimizations
                if issubclass(cls, PeepholeOptimizationStmtBase)
            ]
            self._multistmt_peephole_opts = [
                cls(self.project, self.kb, ail_manager, self.func_addr, self._preserve_vvar_ids, self._type_hints)
                for cls in peephole_optimizations
                if issubclass(cls, PeepholeOptimizationMultiStmtBase)
            ]

        self.result_block = None

        # cached Propagator and ReachingDefinitions results. Clear them if the block is updated
        self._propagator = cached_propagator
        self._reaching_definitions = cached_reaching_definitions

        if self.block is not None:
            self._analyze()

    def _analyze(self):
        block = self.block
        ctr = 0
        max_ctr = 30

        new_block = self._eliminate_self_assignments(block)
        if self._count_nonconstant_statements(new_block) >= 2 and self._has_propagatable_assignments(new_block):
            new_block = self._eliminate_dead_assignments(new_block)
        if new_block != block:
            self._clear_cache()
            block = new_block

        while True:
            ctr += 1
            # block.pp()
            new_block = self._simplify_block_once(block)
            # print()
            # new_block.pp()
            if new_block == block:
                break
            self._clear_cache()
            block = new_block
            if ctr >= max_ctr:
                _l.error(
                    "Simplification does not reach a fixed point after %d iterations. "
                    "Block comparison is probably incorrect.",
                    max_ctr,
                )
                break

        self.result_block = block

    def _compute_propagation(self, block) -> SPropagatorAnalysis:
        if self._propagator is None:
            self._propagator = self.project.analyses[SPropagatorAnalysis].prep(fail_fast=self._fail_fast)(
                subject=block,
                func_addr=self.func_addr,
                stack_pointer_tracker=self._stack_pointer_tracker,
                ail_manager=self._ail_manager,
            )
        return self._propagator

    def _compute_reaching_definitions(self, block) -> SRDAModel:
        if self._reaching_definitions is None:
            self._reaching_definitions = (
                self.project.analyses[SReachingDefinitionsAnalysis]
                .prep(fail_fast=self._fail_fast)(
                    subject=block,
                    track_tmps=True,
                    func_addr=self.func_addr,
                )
                .model
            )
        return self._reaching_definitions

    def _clear_cache(self):
        self._reaching_definitions = None
        self._propagator = None

    @staticmethod
    def _has_propagatable_assignments(block) -> bool:
        return any(isinstance(stmt, (Assignment, Store)) for stmt in block.statements)

    @staticmethod
    def _count_nonconstant_statements(block) -> int:
        return sum(1 for stmt in block.statements if not (isinstance(stmt, Jump) and isinstance(stmt.target, Const)))

    def _simplify_block_once(self, block):
        block = self._peephole_optimize(block)

        nonconstant_stmts = self._count_nonconstant_statements(block)
        has_propagatable_assignments = self._has_propagatable_assignments(block)

        # propagator
        if nonconstant_stmts >= 2 and has_propagatable_assignments:
            propagator = self._compute_propagation(block)
            new_block = block
            if propagator.model is not None:
                replacements = propagator.model.replacements
                if replacements:
                    _, new_block = self._replace_and_build(
                        block, replacements, self._ail_manager, replace_registers=True
                    )
                    new_block = self._eliminate_self_assignments(new_block)
                    self._clear_cache()
        else:
            # Skipped calling Propagator
            new_block = block

        if nonconstant_stmts >= 2 and has_propagatable_assignments:
            new_block = self._eliminate_dead_assignments(new_block)

        return self._peephole_optimize(new_block)

    @staticmethod
    def _replace_and_build(
        block: Block,
        replacements: Mapping[AILCodeLocation, Mapping[Expression, Expression]],
        ail_manager: Manager,
        replace_assignment_dsts: bool = False,
        replace_loads: bool = False,
        gp: int | None = None,
        replace_registers: bool = True,
    ) -> tuple[bool, Block]:
        new_statements = block.statements[::]
        replaced = False

        for codeloc, repls in replacements.items():
            for old, new in repls.items():
                new = new.deep_copy(ail_manager)
                assert codeloc.stmt_idx is not None
                stmt = new_statements[codeloc.stmt_idx]
                if (
                    not replace_loads
                    and isinstance(old, Load)
                    and not isinstance(stmt, SideEffectStatement)
                    and not (gp is not None and isinstance(new, Const) and new.value == gp)
                ):
                    # skip memory-based replacement for non-Call and non-gp-loading statements
                    continue
                if isinstance(stmt, SideEffectStatement) and stmt.expr == old:
                    # the replacement must be a call, since replacements can only be expressions
                    assert isinstance(new, Call)
                    r = True
                    new_stmt = SideEffectStatement(
                        stmt.idx, new, ret_expr=stmt.ret_expr, fp_ret_expr=stmt.fp_ret_expr, **stmt.tags
                    )
                else:
                    # replace the expressions involved in this statement

                    if not replace_registers and isinstance(old, Register):
                        # don't replace
                        r = False
                        new_stmt = None
                    elif isinstance(old, VirtualVariable) and has_reference_to_vvar(stmt, old.varid):
                        # never replace an l-value with an r-value
                        r = False
                        new_stmt = None
                    elif isinstance(stmt, SideEffectStatement) and isinstance(new, Call) and old == stmt.ret_expr:
                        # special case: do not replace the ret_expr of a call statement to another call statement
                        r = False
                        new_stmt = None
                    elif isinstance(stmt, Assignment) and not replace_assignment_dsts:
                        # special case: do not replace the dst
                        new_stmt = None
                        if stmt.src == old:
                            r = True
                            new_src = new.copy()
                        else:
                            r, new_src = stmt.src.replace(old, new)
                        if r:
                            new_stmt = Assignment(stmt.idx, stmt.dst, new_src, **stmt.tags)
                    else:
                        r, new_stmt = stmt.replace(old, new)

                if r:
                    assert new_stmt is not None
                    replaced = True
                    new_statements[codeloc.stmt_idx] = new_stmt

        if not replaced:
            return False, block

        new_block = block.copy()
        new_block.statements = new_statements
        return True, new_block

    @staticmethod
    def _eliminate_self_assignments(block):
        new_statements = []

        for stmt in block.statements:
            if type(stmt) is Assignment:
                if stmt.dst.likes(stmt.src):
                    continue
                if (
                    isinstance(stmt.dst, Register)
                    and isinstance(stmt.src, Convert)
                    and isinstance(stmt.src.operand, Register)
                ) and (
                    stmt.dst.size == stmt.src.size
                    and stmt.dst.reg_offset == stmt.src.operand.reg_offset
                    and not stmt.src.is_signed
                ):
                    # ignore statements like edi = convert(rdi, 32)
                    continue
            new_statements.append(stmt)

        return block.copy(statements=new_statements)

    def _eliminate_dead_assignments(self, block):
        def _statement_has_calls(stmt: Statement) -> bool:
            """
            Check if a statement has any Call expressions.
            """
            walker = HasCallExprWalker()
            walker.walk_statement(stmt)
            return walker.has_call_expr

        def _expression_has_calls(expr: Expression) -> bool:
            """
            Check if an expression has any Call expressions.
            """
            walker = HasCallExprWalker()
            walker.walk_expression(expr)
            return walker.has_call_expr

        new_statements = []
        if not block.statements:
            return block

        rd = self._compute_reaching_definitions(block)
        block_loc = (block.addr, block.idx)

        # Find dead assignments
        dead_defs_stmt_idx = set()
        all_defs = rd.get_all_tmp_definitions(block_loc)
        for d in all_defs:
            assert not d.codeloc.is_extern
            assert not d.dummy

            uses = rd.get_tmp_uses(d.atom, block_loc)
            if not uses:
                dead_defs_stmt_idx.add(d.codeloc.stmt_idx)

        used_tmps: set[int] = set()
        # micro optimization: if all statements that use a tmp are going to be removed, we remove this tmp as well
        for tmp, used_locs in rd.all_tmp_uses[block_loc].items():
            used_at = {stmt_idx for _, stmt_idx in used_locs}
            if used_at.issubset(dead_defs_stmt_idx):  # noqa:SIM102
                # cannot remove this tmp if any use sites involve call expressions; this is basically a duplicate of
                # the logic in the larger loop below
                if all(not _statement_has_calls(block.statements[i]) for i in used_at):
                    continue
            used_tmps.add(tmp.tmp_idx)

        # Remove dead assignments
        for idx, stmt in enumerate(block.statements):
            if type(stmt) is Assignment:
                # tmps can't execute new code
                if (type(stmt.dst) is Tmp and stmt.dst.tmp_idx not in used_tmps) or idx in dead_defs_stmt_idx:
                    # is it assigning to an unused tmp or a dead virgin?

                    # does .src involve any Call expressions? if so, we cannot remove it
                    if not _expression_has_calls(stmt.src):
                        continue

                    if type(stmt.dst) is Tmp and isinstance(stmt.src, Call):
                        # eliminate the assignment and replace it with the call
                        stmt = SideEffectStatement(stmt.idx, stmt.src, **stmt.tags)

                if isinstance(stmt, Assignment) and stmt.src == stmt.dst:
                    continue

            new_statements.append(stmt)

        return block.copy(statements=new_statements)

    #
    # Peephole optimization
    #

    def _peephole_optimize(self, block):
        # expressions are updated in place
        peephole_optimize_exprs(block, self._expr_peephole_opts)

        # run statement-level optimizations
        statements, stmts_updated = peephole_optimize_stmts(block, self._stmt_peephole_opts)

        new_block = block.copy(statements=statements) if stmts_updated else block

        statements, multi_stmts_updated = peephole_optimize_multistmts(new_block, self._multistmt_peephole_opts)

        if not multi_stmts_updated:
            return new_block
        return new_block.copy(statements=statements)


register_analysis(BlockSimplifier, "AILBlockSimplifier")



from __future__ import annotations
import claripy
import logging

from angr.procedures.java import JavaSimProcedure
from angr.engines.soot.values import SimSootValue_StringRef, SimSootValue_ThisRef

log = logging.getLogger(name=__name__)


class IntToInteger(JavaSimProcedure):
    __provides__ = (("java.lang.Integer", "<init>(int)"),)

    def run(self, this_ref, int_val):
        log.debug(f"Called SimProcedure java.lang.Integer.<init> with args: {this_ref} {int_val}")
        this_ref.store_field(self.state, "value", "int", int_val)
        return this_ref


class IntegerToInt(JavaSimProcedure):
    __provides__ = (("java.lang.Integer", "intValue()"),)

    def run(self, this_ref):
        log.debug(f"Called SimProcedure java.lang.Integer.intValue with args: {this_ref}")
        return this_ref.get_field(self.state, "value", "int")


class IntegerToString(JavaSimProcedure):
    __provides__ = (("java.lang.Integer", "toString(int)"),)

    def run(self, int_val):
        log.debug(f"Called SimProcedure java.lang.Integer.toString with args: {int_val}")
        return SimSootValue_StringRef.new_string(self.state, claripy.IntToStr(int_val))


class IntegerValueOf(JavaSimProcedure):
    __provides__ = (("java.lang.Integer", "valueOf(int)"),)

    def run(self, int_val):
        log.debug(f"Called SimProcedure java.lang.Integer.valueOf with args: {int_val}")
        obj_ref = SimSootValue_ThisRef.new_object(self.state, "java.lang.Integer")
        obj_ref.store_field(self.state, "value", "int", int_val)
        return obj_ref



# VEX Variables
from __future__ import annotations


class VEXVariable:
    __slots__ = ()

    def __hash__(self):
        raise NotImplementedError

    def __eq__(self, other):
        raise NotImplementedError


class VEXMemVar:
    __slots__ = (
        "addr",
        "size",
    )

    def __init__(self, addr: int, size: int):
        self.addr = addr
        self.size = size

    def __hash__(self):
        return hash((VEXMemVar, self.addr, self.size))

    def __eq__(self, other):
        return type(other) is VEXMemVar and other.addr == self.addr and other.size == self.size

    def __repr__(self):
        return f"<mem {self.addr:#x}[{self.size} bytes]>"


class VEXReg(VEXVariable):
    __slots__ = (
        "offset",
        "size",
    )

    def __init__(self, offset, size):
        self.offset = offset
        self.size = size

    def __hash__(self):
        return hash((VEXReg, self.offset, self.size))

    def __eq__(self, other):
        return type(other) is VEXReg and other.offset == self.offset and other.size == self.size

    def __repr__(self):
        return f"<reg {self.offset}[{self.size}]>"


class VEXTmp(VEXVariable):
    __slots__ = ("tmp",)

    def __init__(self, tmp):
        self.tmp = tmp

    def __hash__(self):
        return hash((VEXTmp, self.tmp))

    def __eq__(self, other):
        return type(other) is VEXTmp and other.tmp == self.tmp

    def __repr__(self):
        return f"<tmp {self.tmp}>"



from __future__ import annotations
from angr.sim_type import SimTypeFunction, SimTypePointer, SimTypeChar, SimTypeBottom
from angr.procedures import SIM_PROCEDURES as P
from . import SimCppLibrary

libstdcpp = SimCppLibrary()

libstdcpp.set_library_names("libstdc++.so", "libstdc++.so.6")
libstdcpp.add_all_from_dict(P["libstdcpp"])


_decls = {
    "std::__throw_logic_error(char const*)": SimTypeFunction(
        [SimTypePointer(SimTypeChar())], SimTypeBottom(label="void"), arg_names=("error",)
    ),
    "std::__throw_length_error(char const*)": SimTypeFunction(
        [SimTypePointer(SimTypeChar())], SimTypeBottom(label="void"), arg_names=("error",)
    ),
}


for name, proto in _decls.items():
    if proto is not None:
        libstdcpp.set_prototype(name, proto)



from __future__ import annotations
import claripy
from angr import ailment
from angr.state_plugins.callstack import CallStack
from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class AILCallStack(CallStack):
    """
    An implementation of state.callstack for AIL symbolic execution

    Contains extra state for local vars and AIL calling conventions
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.passed_args: tuple[claripy.ast.Bits, ...] | None = None
        self.passed_rets: tuple[tuple[claripy.ast.Bits, ...], ...] = ()
        self.resume_at: int | None = None
        self.vars: dict[int, claripy.ast.Bits | MemoryMixin] = {}
        self.var_refs: dict[claripy.ast.BV, int] = {}
        self.var_refs_rev: dict[int, claripy.ast.BV] = {}
        self.return_addr: ailment.Address | None = None

    @CallStack.memo
    def copy(self, memo):
        o = super().copy(memo)
        o.passed_args = self.passed_args
        o.passed_rets = self.passed_rets
        o.resume_at = self.resume_at
        o.vars = {idx: v.copy(memo) if isinstance(v, MemoryMixin) else v for idx, v in self.vars.items()}
        o.var_refs = dict(self.var_refs)
        o.var_refs_rev = dict(self.var_refs_rev)
        o.return_addr = self.return_addr
        return o

    @CallStack.memo
    def copy_without_tail(self, memo):
        o = super().copy_without_tail(memo)
        o.passed_args = self.passed_args
        o.passed_rets = self.passed_rets
        o.resume_at = self.resume_at
        o.vars = {idx: v.copy(memo) if isinstance(v, MemoryMixin) else v for idx, v in self.vars.items()}
        o.var_refs = dict(self.var_refs)
        o.var_refs_rev = dict(self.var_refs_rev)
        o.return_addr = self.return_addr
        return o

    def set_state(self, state):
        super().set_state(state)
        for val in self.vars.values():
            if isinstance(val, MemoryMixin):
                val.set_state(state)

    def _manage(self):
        # this is managed by the engine
        pass



from __future__ import annotations

from angr.ailment.statement import Statement
from angr.ailment.expression import VirtualVariable
from angr.ailment.block import Block

from angr.code_location import AILCodeLocation


class RewritingState:
    """
    The abstract state for the expression rewriting engine.
    """

    def __init__(
        self,
        loc: AILCodeLocation,
        arch,
        func,
        original_block: Block,
        registers: dict[int, VirtualVariable] | None = None,
        stackvars: dict[int, VirtualVariable] | None = None,
    ):
        self.loc = loc
        self.arch = arch
        self.func = func

        self.registers = registers or {}
        self.stackvars = stackvars or {}
        self.tmps: dict[int, VirtualVariable] = {}
        self.original_block = original_block
        self.out_block = None

    def copy(self) -> RewritingState:
        # copy_regs = defaultdict(dict)
        # for k, vdict in self.registers.items():
        #     copy_regs[k] = vdict.copy()

        copy_regs = dict(self.registers)
        copy_stackvars = dict(self.stackvars)
        # copy_stackvars = defaultdict(dict)
        # for k, vdict in self.stackvars.items():
        #     copy_stackvars[k] = vdict.copy()

        return RewritingState(
            self.loc,
            self.arch,
            self.func,
            self.original_block,
            registers=copy_regs,
            stackvars=copy_stackvars,
        )

    def append_statement(self, stmt: Statement):
        if self.out_block is None:
            self.out_block = Block(self.loc.block_addr, self.original_block.original_size, idx=self.loc.block_idx)
        self.out_block.statements.append(stmt)



from __future__ import annotations
import logging

from .base import SimSootStmt

l = logging.getLogger("angr.engines.soot.statements.invoke")


class SimSootStmt_Invoke(SimSootStmt):
    def _execute(self):
        invoke_expr = self._translate_expr(self.stmt.invoke_expr)
        self._add_invoke_target(invoke_expr)



from __future__ import annotations

import logging

import cle
from cle.structs import DataDirectory, MemRegion, MemRegionSort

from angr.knowledge_plugins.cfg.memory_data import MemoryDataSort

log = logging.getLogger(name=__name__)

# Mapping from MemRegionSort to the MemoryDataSort used in CFGFast's _seg_list and memory_data
_SORT_MAP: dict[MemRegionSort, str | None] = {
    MemRegionSort.IAT: MemoryDataSort.PointerArray,
    MemRegionSort.ILT: MemoryDataSort.PointerArray,
    MemRegionSort.EXPORT_ADDR_TABLE: MemoryDataSort.PointerArray,
    MemRegionSort.EXPORT_NAME_TABLE: MemoryDataSort.PointerArray,
    MemRegionSort.EXPORT_ORDINAL_TABLE: MemoryDataSort.Integer,
    MemRegionSort.EXPORT_DIRECTORY: MemoryDataSort.PEExportDirectory,
    MemRegionSort.IMPORT_DIRECTORY: MemoryDataSort.PEImportDirectory,
    MemRegionSort.IMPORT_HINT_NAME_TABLE: MemoryDataSort.String,
    MemRegionSort.DELAY_IMPORT_DIRECTORY: MemoryDataSort.PEDelayImportDirectory,
    MemRegionSort.STRING_BLOB: MemoryDataSort.String,
    MemRegionSort.POINTER_ARRAY: MemoryDataSort.PointerArray,
    MemRegionSort.STRUCT_ARRAY: MemoryDataSort.Unknown,
    MemRegionSort.DATA: MemoryDataSort.Unknown,
}


def _flatten_regions(regions: list[MemRegion]) -> list[MemRegion]:
    """Flatten DataDirectory regions into their sub-regions."""
    result = []
    for region in regions:
        if isinstance(region, DataDirectory) and region.sub_regions:
            result.extend(region.flat_regions())
        else:
            result.append(region)
    return result


def get_data_regions_from_meta_regions(loader: cle.Loader) -> list[tuple[int, int, str | None]]:
    """
    Extract (addr, size, sort) tuples for all metadata data regions across all loaded objects.

    Flattens DataDirectory sub-regions so each entry maps to a contiguous memory range
    suitable for marking in CFGFast's _seg_list.
    """
    result = []
    for obj in loader.all_objects:
        if not hasattr(obj, "meta_regions") or not obj.meta_regions:
            continue

        flat = _flatten_regions(obj.meta_regions)
        for region in flat:
            sort = _SORT_MAP.get(region.sort, MemoryDataSort.Unknown)
            if region.size > 0:
                result.append((region.vaddr, region.size, sort))

    return result


def get_function_hints_from_meta_regions(loader: cle.Loader) -> list[tuple[int, str | None]]:
    """
    Extract (addr, name) pairs for functions discovered from metadata.

    Uses meta_function_hints populated by backends (e.g., PE export table entries).
    """
    result = []
    for obj in loader.all_objects:
        if hasattr(obj, "meta_function_hints") and obj.meta_function_hints:
            result.extend(obj.meta_function_hints)
    return result



from __future__ import annotations
import angr


class strndup(angr.SimProcedure):
    # pylint:disable=arguments-differ, missing-class-docstring

    def run(self, s, n):
        strnlen = angr.SIM_PROCEDURES["libc"]["strnlen"]
        memcpy = angr.SIM_PROCEDURES["libc"]["memcpy"]
        malloc = angr.SIM_PROCEDURES["libc"]["malloc"]

        src_len = self.inline_call(strnlen, s, n).ret_expr
        new_s = self.inline_call(malloc, src_len + 1).ret_expr

        self.inline_call(memcpy, new_s, s, src_len)

        return new_s
